/*
 * Copyright (C) 2007        Rildo Pragana, David Essex.
 * Copyright (C) 1999-2004,  Rildo Pragana, Ferran Pegueroles, Jim Noeth,
 *                           Bernard Giroud, David Essex, Andrew Cameron. 
 * Copyright (C) 1993, 1991  Rildo Pragana.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1,
 * or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; see the file COPYING.LIB.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA 02111-1307 USA
  
 * 
 * TinyCOBOL Run Time Library -- File handling functions
 * 

*/

#include "htcoblib.h"
#include "fileio.h"

//#define DEBUG_FILEIO_RTS0 1  /* Debug open and close files */
//#define DEBUG_FILEIO_RTS1 1  /* Debug sequential files */
//#define DEBUG_FILEIO_RTS2 1  /* Debug line-sequential files */
//#define DEBUG_FILEIO_RTS3 1  /* Debug relative files */

#ifdef  USE_MF_COMPATABILITY
#define TCOB_EXTERNFILE_MFCOMP  1
#endif

#define bcounter         5

int ignore_lock=0;
int need_lock=0;


/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          save_status                                   |
 |                                                                        |
\*------------------------------------------------------------------------*/
//static void save_status (struct file_desc *f, int status)
void save_status (struct file_desc *f, int status)
{
  f->file_status[0] = status / 10 + '0';
  f->file_status[1] = status % 10 + '0';
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_save_status                              |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_save_status(char *status, int rt) 
{
  status[0] = rt/10+'0';
  status[1] = rt%10+'0';
  return rt;
}

#ifdef TCOB_EXTERNFILE_MFCOMP 

/*-----------------------------------------------------------*\
 *  map_file_name                                            *
 *                                                           *
 *  When the programmer wrote the COBOL SELECT statement,    *
 *  we assume he intended the User to define an Environment  *
 *  Variable that will tell the program where to find the    *
 *  file.                                                    *
 *                                                           *
 *  There are several possibilities:                         *
 *                                                           *
 *  1.  The programmer wants the User to specify the base    *
 *      directory for the data.  In this case, the           *
 *      programmer would have put something like             *
 *      "DATA/Stock/ProductFile" in the SELECT clause.       *
 *                                                           *
 *      In this case, we would expect the User to set an     *
 *      Environment Variable named "$DD_DATA".               *
 *      Example (i)                                          *
 *          DD_DATA=/usr/local/AccountsData                  *
 *          In this case, the full name of the file will     *
 *          be: /usr/local/AccountsData/Stock/ProductFile    *
 *                                                           *
 *      However, Users like to do dumb things so we also     *
 *      check to see whether they set up an environment      *
 *      Variable named "$DATA".                              *
 *      Example (ii)                                         *
 *          DATA=/u/Accts                                    *
 *          In this case, the full name of the file will     *
 *          be :  /u/Accts/Stock/ProductFile                 *
 *                                                           *
 *      (Note that these two examples prove that dumb Users  *
 *      always use shorter, less meaningful names than the   *
 *      theoretically possible intellegent User.)            *
 *                                                           *
 *  2.  The programmer wants the User to specify the full    *
 *      file name.  In this case, the programmer would have  *
 *      put something like "PRODUCTFILE" in his SELECT       *
 *      statement.                                           *
 *                                                           *
 *      In this case, we would like the User to have set     *
 *      up an Environment Variable named "$DD_PRODUCTFILE".  *
 *      Example (iii)                                        *
 *          DD_PRODUCTFILE=/u/Data/Stock/ProductFile         *
 *          In this case, the full name of the file will     *
 *          simply be : /u/Data/Stock/ProductFile            *
 *                                                           *
 *      Again, we need to cater for the less intelligent     *
 *      User, so we also check to see whether he named the   *
 *      Environment Variable "$PRODUCTFILE".                 *
 *      Example (iv)                                         *
 *          PRODUCTFILE=/ProductFile                         *
 *          In this case, the full name of the file will     *
 *          simply be : /ProductFile                         *
 *      (Note that example (iv) is a true to life example    *
 *      showing just how dumb Users can be.  It takes a      *
 *      truly ingenious kind of studidity to manage to       *
 *      set your data files up in the root directory.)       *
 *                                                           *
 *  3.  (a) The programmer wants to hard-code the name of    *
 *          file himself and hence will have something       *
 *          like "/Data/Stock/ProductFile" in his            *
 *          SELECT statement.                                *
 *      (b) The programmer really wanted the User to         *
 *          specify an Environment Variable, but the User    *
 *          stuffed up the installation process and there    *
 *          is no Environment Variable to be found.          *
 *                                                           *
 *      In this case, we can do nothing except to use the    *
 *      name the programmer typed in the SELECT statement    *
 *      as the actual name of the file.                      *
 *      That is, we actually look for a file called :        *
 *      "/Data/Stock/ProductFile".                           *
 *                                                           *
 *      (Note that in this case we can be confident that     *
 *      "/Data" is a separate partition and not a            *
 *      directory in the root partition - Yeah, sure!)       *
 *                                                           *
\*-----------------------------------------------------------*/
char *map_file_name(char *evname)
{
        int slash_pos;
        int len, lenDD, lenSuffix;
        char *file_name;
        char *evDataDir, *dataDir, *file_Suffix;

        file_name = NULL;

//        fprintf(stderr, "debug: map_file_name: LT_DIRSEP_STR=%s;\n", LT_DIRSEP_STR);  

        /* First we search through the file name specified
         * by the COBOL programmer and find the first
         * occurance of the character "/".
         * We will then use this to divide the name into
         * the "Data Directory" and the "File Suffix".
         * ( If there is no "/" in the name, then we treat
         * the entire name as the "Data Directory" and
         * assume a NULL File Suffix.  I know it sounds
         * crazy, but no less crazy than assuming a Null
         * Data Directory). 
         */
        slash_pos = strcspn(evname, LT_DIRSEP_STR);

        /* Now we cheat a little by setting file_suffix
         * to point to the same string as "evname",
         * but starting half way through.
         * This is naughty, and we must be careful that
         * we don't go changing the contents of evname.
         */
        file_Suffix = evname + slash_pos;
        lenSuffix = strlen(file_Suffix);

        /* Next we start forming up the name of the 
         * Environment Variable we wepect to find.
         * We allocate enough room to hold the
         * prefix for the environment variable 
         *  :-     "DD_"
         * as well as the "Name" and the terminating
         * Null character. 
         */
        evDataDir = malloc(slash_pos + 4);

        strncpy(evDataDir, "DD_", 3);
        strncpy(evDataDir + 3, evname, slash_pos);
        evDataDir[slash_pos + 3] = 0;

        /* Now lets see if the Environment Variable 
         * has been set up/
         */
        dataDir = getenv(evDataDir);

        /*  If we failed to find an environment variable named 
         *  "$DD_DataDir" then we don't give up yet...
         *
         *  We will see whether the User has defined an
         *  environment variable named "$DataDir".
         *  Obviously, To do this, we just drop the first three
         *  characters from "DD_DataDir".
         */
        if (dataDir == NULL)
                {
                evDataDir = evDataDir + 3;
                dataDir = getenv(evDataDir);
                }

        /*  If we found an environment variable containing the
         *  Data Directory, then we go ahead and use it.
         *  That is, we glue the file Suffix onto the end
         *  of the Data Directory to give us the full name
         *  of the file.
         */
        if (dataDir != NULL)
                {
                lenDD = strlen(dataDir);
                file_name = malloc(lenDD + lenSuffix + 1);
                strcpy(file_name, dataDir);
                strcpy(file_name + lenDD, file_Suffix);
                }

        else
                /*  If we failed to find an environment variable
                 *  containing the name of the Data Directory,
                 *  then we assume that the Programmer has
                 *  Hard-Coded the file name.
                 *  That is, we assume the name of the Environment
                 *  Variable is really the name of the file itself.
                 */
                {
                len = strlen(evname);
                file_name = malloc(len+1);
                strncpy(file_name,evname,len);
                file_name[len] = '\0';
                }

/*        fprintf(stderr, "debug: map_file_name: file_name=%s;\n", file_name); */

        return(file_name);

} 
#endif


/*------------------------------------------------------------------------*\
 |                                                                        |
 |                    tcob_resolve_file_name                              |
 |                                                                        |
\*------------------------------------------------------------------------*/
char *tcob_resolve_file_name(struct file_desc *f, char *fname)
{
  int len;
  char *filename=NULL, *evname=NULL, *pt;

  /*
  BEWARE: 
  fname points to a field storage (non null-terminated) 
  we must copy fname to a C string and terminate it with a '\0'
  and also trim spaces at the end. 

  NOTE: 
  The 'file_external' flag in the file structure 'struct file_desc'
  is used to indicate how the value stored in 'fname' is used.
  It can be used to define a filename variable or a environment
  variable.
  
  If the flag is set, then 'fname' is used as an environment 
  variable which is used to determined the actual filename.
  Else 'fname' is used to define a filename. 
  
  If no environment variable is found or is empty, then the 
  filename defaults to the value stored in 'fname'.
  
  Standard input/output/error files have the 'file_external' flag
  set on by default.
  The environment variables STANDARD_[INPUT | OUTPUT | ERROR] can be used 
  to override (redirect) these files.
  
  If an environment variable is found and is not empty, then the name in the 
  environment variable is used as a file name.

  If no environment variable is found or is empty, then these files defualt to
  standard input/output/error as defined by the originating program.
  
  
  */

//#ifdef DEBUG_FILEIO_RTS0
// fprintf(stderr, "debug : tcob_resolve_file_name 0260.020 : filename=%s, reclen=%d;\n", filename, f->reclen);
//#endif                                       

  if (f->flags.file_external == 0) {
     len = f->fname_desc->len;
     filename = malloc(len + 1);
     memmove(filename, fname, len);
     do {
        filename[len--] = 0;
     } while (filename[len] == ' ');
     evname = NULL;
  } 
  else {
//     len = f->fname_desc->len;
//     evname = malloc(len + 1);
     if ((f->flags.file_stdin == 1)  ||
         (f->flags.file_stdout == 1) ||
         (f->flags.file_stderr == 1))
     {
        len = 23;
        evname = malloc(len + 1);
        memset(evname, '\0', len + 1);
        if (f->flags.file_stdin == 1)
            strncpy(evname, "STANDARD_INPUT", len);
        if (f->flags.file_stdout == 1)
            strncpy(evname, "STANDARD_OUTPUT", len);
        if (f->flags.file_stderr == 1)
            strncpy(evname, "STANDARD_ERROR", len);
     }
     else {
        len = f->fname_desc->len;
        evname = malloc(len + 1);
        memmove(evname, fname, len);
        do {
            evname[len--] = 0;
        } while (evname[len] == ' ');
     }

     /* Get environment variable, if it exists */
#ifdef TCOB_EXTERNFILE_MFCOMP 
     filename = map_file_name(evname);
#else
     if ((pt = getenv(evname)) != NULL) {
        len = strlen(pt);
        if (len > 0) {
           filename = malloc(len + 1);
           strncpy(filename, pt, len);
           /* Override standard input/output/error files by environment variable*/
           f->flags.file_stdin  = 0;
           f->flags.file_stdout = 0;
           f->flags.file_stderr = 0;
        } 
        else {
           len = strlen(evname);
           filename = malloc(len + 1);
           strncpy(filename, evname, len);
        }
     } 
     else {
        len = strlen(evname);
        filename = malloc(len + 1);
        strncpy(filename, evname, len);
     }
     filename[len] = '\0';
#endif
  }

  if (evname != NULL)
     free(evname);

#ifdef DEBUG_FILEIO_RTS0
 fprintf(stderr, "debug : tcob_resolve_file_name 0260.120 : filename=%s, len=%d;\n", filename, len);
#endif                                       
  
  return (filename);
} 

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_open                                     |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_open(struct file_desc *f, char *record, char *fname, int mode) 
{
 int r;

 /* Check the correct structure */
 if (f->vers_id < RTL_FILE_VERSION) {
     fprintf(stderr, "You need to recompile your program\n");
     fprintf(stderr, "Version mismatch; structure %x, RTL %x\n",
             f->vers_id, RTL_FILE_VERSION);
     RETURN_STATUS(99);
 }

 r = tcob_open_real(f, fname, mode);

 /* save the file in the list of open files for the proces */
 if (r == 0) 
    tcob_add_file_list(f);

 RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_open_real                                |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_open_real(struct file_desc *f, char *fname, int mode) 
{
 int sflags = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH; 
 int oflags=0;
 char *filename;
 unsigned char file_organization;
 int r=0;
 
 /* Check to see if the file is already open. If so return
    File Status 41 in according to the Ansi 85 Standard. */
 if (f->dbp != NULL)
     RETURN_STATUS(41);

 f->flags.read_done = 0;
 f->recloc = -1;
 f->flags.eof_hit = 0;

#ifdef DEBUG_FILEIO_RTS0
 fprintf(stderr, "debug : tcob_open 0370.010 : f->flags file_external=%d, eof_hit=%d, file_printer=%d, file_stdin=%d, file_stdout=%d, file_stderr=%d;\n", 
                        f->flags.file_external, f->flags.eof_hit, f->flags.file_printer, f->flags.file_stdin, f->flags.file_stdout, f->flags.file_stderr);
#endif                                       

 filename = tcob_resolve_file_name(f, fname);
#ifdef DEBUG_FILEIO_RTS0
 fprintf(stderr, "debug : tcob_open 0370.020 : filename=%s, reclen=%d;\n", filename, f->reclen);
#endif                                       
 if (filename == NULL) 
    return (99);
          
 switch (mode) {

   case FMOD_INPUT: 
        /* Standard output type files do not support INPUT mode */
        if ((f->flags.file_stderr == 1) ||
            (f->flags.file_stdout == 1))
        {
           return (99);
        }
        oflags = O_RDONLY; 
        break;
   case FMOD_IO: 
        if (f->organization == ORG_LINESEQUENTIAL) {
        /* Line Sequential does not Support Mode IO */
           return (37);
        }
        /* Standard input/output/error type files do not support IO mode */
        if ((f->flags.file_stdin == 1)  ||
            (f->flags.file_stdout == 1) ||
            (f->flags.file_stderr == 1))
        {
           return (99);
        }
        oflags = O_RDWR;
        if (f->flags.optional) 
           oflags = O_CREAT | O_RDWR;
        break;
   case FMOD_OUTPUT: 
        /* Standard input type files don't support OUTPUT mode */
        if (f->flags.file_stdin == 1) {
           return (99);
        }
        /* BDB doesn't support O_WRONLY */ 
        oflags = O_CREAT | O_TRUNC | O_RDWR; 
        break;
   case FMOD_EXTEND: 
        /* Standard input type files don't support OUTPUT mode */
        if (f->flags.file_stdin == 1) {
           return (99);
        }
        oflags =  O_RDWR | O_APPEND;
        if (f->flags.optional)
            oflags = O_CREAT | O_RDWR | O_APPEND;
        break;
 }

#if defined(__MINGW32__) || defined(__CYGWIN__)
 oflags |= O_BINARY;
#endif

#ifdef USE_SYNC
 if (f->organization == ORG_INDEXED)
    oflags |= O_SYNC;
#endif

 /* Initialize flags */
 f->flags.start_record = 0; 
 f->recloc = -1;
 f->flags.read_done = 0;

#ifdef DEBUG_FILEIO_RTS0
 fprintf(stderr, "debug : tcob_open 0370.040 : filename=%s, f->fname_desc->len=%d, mode=%d, oflags=%d, sflags=%d;\n", filename, f->fname_desc->len, mode, oflags, sflags);
#endif                                       

 /* A printer SEQUENTIAL file is equivalent to a LINE SEQUENTIAL file */
 if ((f->flags.file_printer == 1) && (f->organization == ORG_SEQUENTIAL)) 
    file_organization = ORG_LINESEQUENTIAL;
 else
    file_organization = f->organization;
 
 switch (file_organization)
 {
   case ORG_INDEXED:
     r = tcob_open_idx(f, filename, mode, oflags, sflags); 

     if (r != 0)
        f->dbp = NULL;
                 
     free(filename);
     return r; 
     break;

   case ORG_RELATIVE:
     if ((int)f->dbp == -1) {
        f->dbp=0;
     }
     else {          
        /* File IO record buffers */ 
        f->sbuf1 = (char *)malloc((sizeof(char) * f->reclen) + 5);
        /* File IO key buffers */ 
        f->kbuf1 = (char *)malloc((sizeof(char) * f->reclen) + 5);
        f->kbuf2 = (char *)malloc((sizeof(char) * f->reclen) + 5);
        f->dbp = (void *)open(filename, oflags, sflags);
     }
     break;

   case ORG_LINESEQUENTIAL:
     if (mode == FMOD_INPUT) {
        if (f->flags.file_stdin == 1) 
           f->dbp = stdin;
        else 
           f->dbp = (void *)fopen(filename, "r");
     }
     else {
        if ((f->flags.file_stderr == 1) ||
            (f->flags.file_stdout == 1))
        {
           if (f->flags.file_stdout == 1)
              f->dbp = stdout;
           if (f->flags.file_stderr == 1)
              f->dbp = stderr;
        }
        else {
           f->dbp = (void *)fopen(filename, "w");
           /* f->dbp = (void *)open(filename, oflags, sflags); */
        }
     }
     if (f->dbp != NULL) {
        /* File IO record buffers */ 
        f->sbuf1 = (char *)malloc((sizeof(char) * f->reclen) * 2);
     }
     break;
   
   /* Otherwise it is a sequential file */
   default:
     if ((f->flags.file_stdin == 1)  ||
         (f->flags.file_stdout == 1) ||
         (f->flags.file_stderr == 1))
     {
        if (f->flags.file_stdin == 1)
           f->dbp = (void *)0;
        if (f->flags.file_stdout == 1)
           f->dbp = (void *)1;
        if (f->flags.file_stderr == 1)
           f->dbp = (void *)2;
     }
     else {
        f->dbp = (void *)open(filename, oflags, sflags);
     }
     if ((int)f->dbp == -1) {
        f->dbp=0;
     }
     else {
        /* File IO record buffers */ 
        f->sbuf1 = (char *)malloc((sizeof(char) * f->reclen) * 2);
     }
     break;
 }
 
 if (f->dbp == NULL) {
     /* Check to see if optional */
     if (f->flags.optional && mode == FMOD_INPUT) {
        f->dbp = NULL;
        f->flags.file_missing = 1;
     } 
     else {
        if (errno == EINVAL) {
           f->dbp = NULL; 
           free(filename);
           return (37); 
        }
        if (errno == ENOENT) {
           f->dbp = NULL; 
           free(filename);
           return (35);
        }
        f->dbp = NULL; 
        free(filename); 
        return (91);
     }
 }

 /* save mode to check later (read,write,start,...) */
 f->open_mode = mode;

 return r;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                            tcob_close                                  |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_close( struct file_desc *f, char *record) 
{
 int r;

 /* Close file and free buffers if requird */
 r = tcob_close_real(f, record);
 /* Remove file from Run-Time open files list */
 if (r == 0) 
    tcob_remove_file_list(f);

 RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_close_real                               |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_close_real(struct file_desc *f, char *record) 
{
  int r=0;
  unsigned char file_organization;

  f->flags.read_done = 0;
  
  /* Check to see if file is open. 
   If not return File Status 42 
   In accordance with the COBOL 85 Standard.  */
  if (f->dbp == NULL) {
     if (f->flags.optional && f->flags.file_missing) 
         return (0);
     return (42);
  }

  /* A printer SEQUENTIAL file is equivalent to a LINE SEQUENTIAL file */
  if ((f->flags.file_printer == 1) && (f->organization == ORG_SEQUENTIAL)) 
     file_organization = ORG_LINESEQUENTIAL;
  else
     file_organization = f->organization;
  
  switch (file_organization)
  {
    case ORG_INDEXED:
      r = tcob_close_idx(f, record);
      break;    
    
    case ORG_RELATIVE:
      close((int)f->dbp);
      f->dbp = NULL;

      if (f->sbuf1 != NULL) 
         free(f->sbuf1);
      if (f->kbuf1 != NULL) 
         free(f->kbuf1);
      if (f->kbuf2 != NULL) 
         free(f->kbuf2);
     
      break;

    case ORG_LINESEQUENTIAL:
      if ((f->flags.file_stdin == 0)  &&
          (f->flags.file_stdout == 0) &&
          (f->flags.file_stderr == 0))
      {
         fclose((FILE *)f->dbp);
      }
      break;
    
    default:
      if ((f->flags.file_stdin == 0)  &&
          (f->flags.file_stdout == 0) &&
          (f->flags.file_stderr == 0))
      {
         close((int)f->dbp);
      }
      break;
  }
    
  f->dbp = NULL;
  
  return r;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_file_init                                |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_file_init( struct file_desc *f, char *record) 
{
#ifdef DEBUG_RTS
        fprintf(stderr, "file_init: \n");
#endif
        tcob_close(f, record);
        f->open_mode = 0;
        f->flags.file_missing = 0;
        return 0;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_setlock                                  |
 |                                                                        |
\*------------------------------------------------------------------------*/
void tcob_setlock() 
{
  /* this just make aware we need to lock current record, used before a read */
  need_lock = 1;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                       tcob_lock_record                                 |
 |                                                                        |
\*------------------------------------------------------------------------*/
/* static int tcob_lock_record(int fd, int len) */
int tcob_lock_record(int fd, int len) 
{
#if defined(__MINGW32__)
   return -1;      
#else
#ifdef HAVE_FCNTL
   struct flock lk;
   lk.l_type = F_WRLCK;
   lk.l_whence = SEEK_CUR;
   lk.l_start = 0;
   lk.l_len = len;
   lk.l_pid = getpid();
   need_lock = 0;
   return (fcntl(fd,F_SETLK,&lk));        
#else
   return -1;      
#endif
#endif
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                        tcob_unlock_record                              |
 |                                                                        |
\*------------------------------------------------------------------------*/
void tcob_unlock_record(struct file_desc *f) 
{
#if defined(__MINGW32__)
   return; 
#else
#ifdef HAVE_FCNTL
   {
   struct flock lk;
   lk.l_type = F_UNLCK;
   lk.l_whence = SEEK_CUR;
   lk.l_start = 0;
   lk.l_len = f->reclen;
   lk.l_pid = getpid();
   fcntl((int)f->dbp,F_SETLK,&lk);
   }        
#else
   return; 
#endif
#endif
}


/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_ignorelock                               |
 |                                                                        |
\*------------------------------------------------------------------------*/
void tcob_ignorelock() 
{
 /* this just make aware we need to ignore locks on the current record, used before a read */
        ignore_lock = 1;
}

//+---------------------------------------------------------------------------------+
// Line-sequential are variable length files which use a <LF> as a record delimiter.                                                        
//                                                       
//+---------------------------------------------------------------------------------+
/*------------------------------------------------------------------------*\
 |                                                                        |
 |                      tcob_read_lseq                                    |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_read_lseq(struct file_desc *f, char *record, struct fld_desc *reclen_desc, char *reclen_buf) 
{
        int reclen, r=0;
        int ich;
        unsigned char *ptrec;
        FILE *fdbp;

        f->flags.read_done = 0;
        /* Check to see if file is open. 
           If not return File Status 47
           in accordance with the COBOL 85 Standard. */
        if (f->dbp == NULL) {
            if (f->flags.optional && f->flags.file_missing) 
               RETURN_STATUS(10);        
            RETURN_STATUS(47);
        }

        /* Check to see that the record length is valid */
        if (f->reclen == -1)
            RETURN_STATUS(99);
        
        /* Check the mode the file is opened in to make sure that read
           is Allowed */
        if (((f->open_mode != FMOD_INPUT) && (f->open_mode != FMOD_IO)))
           RETURN_STATUS(47);
        
        /* Check that we didn't already hit eof */
        if ((f->open_mode == FMOD_INPUT || f->open_mode == FMOD_IO) && f->flags.eof_hit)
           RETURN_STATUS(46);
        
#ifdef DEBUG_FILEIO_RTS2
        fprintf(stderr, "debug : tcob_read_lseq 0945.060 : f->reclen=%d\n", f->reclen);
#endif                 

        fdbp = (FILE *)f->dbp;
        memset(record, ' ', f->reclen);
        reclen = 0;
        ptrec = record;
        ich = 0;
        
        ich = fgetc(fdbp);
        while (ich != EOF) { 
          if (ich == '\n') {
             break;
	  }
          if (ich != '\r') {
             if(reclen < f->reclen) {
               *ptrec++ = (char)ich;
               reclen++;
             }
	  }
          ich = fgetc(fdbp);
        }

        // Copy variable length data  
        if (reclen_desc != NULL)                 
           tcob_move(&_generic_4binary, (char *)&reclen, reclen_desc, reclen_buf);             

#ifdef DEBUG_FILEIO_RTS2
        fprintf(stderr, "debug : tcob_read_lseq 0945.120 : reclen=%d, f->reclen=%d\n", reclen, f->reclen);
#endif                 

        if (reclen > f->reclen)
            RETURN_STATUS(99);
        if (ich == EOF)
           RETURN_STATUS(10);

        f->flags.read_done = 1;

        RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                       tcob_write_lseq                                  |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_write_lseq(struct file_desc *f, char *record, struct fld_desc *reclen_desc, char *reclen_buf) 
{
        int reclen, r=0, i;
        FILE *fdbp;
        
        f->flags.read_done = 0;
        
        /* Check to see if file is open. 
           If not return File Status 48
           In accordance with the COBOL 85 standard. */
        if (f->dbp == NULL)
           RETURN_STATUS(48);
        
        /* Check to see that the record length is valid */
        if (f->reclen == -1) {
           RETURN_STATUS(99);
         }
       
#ifdef DEBUG_FILEIO_RTS2
         fprintf(stderr, "debug : tcob_write_lseq 1200.080 : f->reclen=%d\n\n", f->reclen);
#endif
        /* 
         Check the mode the file is opened in to make sure that write is Allowed 
        */
        if (((f->open_mode != FMOD_OUTPUT) && 
             (f->open_mode != FMOD_IO) && 
             (f->open_mode != FMOD_EXTEND)))
           RETURN_STATUS(48);

        fdbp = (FILE *)f->dbp;

        // If variable length record is defined 
        if (reclen_desc != NULL) {

           // Convert record to int type 
           tcob_move(reclen_desc, reclen_buf, &_generic_4binary, (char *)&reclen);     
#ifdef DEBUG_FILEIO_RTS2
           fprintf(stderr, "debug : tcob_write_lseq 1200.180 : reclen=%d, f->reclen=%d\n", reclen, f->reclen); 
#endif     
      
           // Check that the record length is not greater than max  
           if (reclen > f->reclen)
              RETURN_STATUS(99);

           // Write the record with no trailing blanks truncation 
           for (i=0; i<reclen; i++) {
             fputc(*(record + i), fdbp);
           }
           // Write the record delimiter 
           // A printer SEQUENTIAL file is equivalent to a LINE SEQUENTIAL file 
           if (f->flags.file_printer == 1) {
              fputc('\r', fdbp);
              fputc('\n', fdbp);
           } 
           else {
              fputc('\n', fdbp);
           } 
        } 
        else {
           // 
           //  Standard LS record with trailing blanks truncation 
           //  Determine length of record by truncating trailing blanks 
           //
           reclen = f->reclen;
           if (f->flags.file_printer == 0) {
              for(i=f->reclen - 1; i >= 0; i--) {
                  if (*(record + i) != ' ') {
                     reclen = i + 1;
//                     i = -1;
                     break;
                  }
              }
           } 
           // Write the record
           if(reclen > 0) {
             for (i=0; i<reclen; i++) {
               fputc(*(record + i), fdbp);
             }
           } 
           // Write the record delimiter  
           if (f->flags.file_printer == 1) {
              fputc('\r', fdbp);
              fputc('\n', fdbp);
           } 
           else {
             fputc('\n', fdbp);
           } 
        } 
                 
        RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                       tcob_write_adv_lseq                              |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_write_adv_lseq(struct file_desc *f, char *record, int opt, 
                        struct fld_desc *cnt_desc, char *cnt_buf, 
                        struct fld_desc *reclen_desc, char *reclen_buf) 
{
        int i, r=0, len=0, lines=0;
        FILE *fdbp; 

        /* Check to see if file is open. 
           If not return File Status 48
           In accordance with the COBOL 85 Standard. */
        if (f->dbp == NULL)
           RETURN_STATUS(48);
        
         fdbp = (FILE *)f->dbp;
        
        /* Check the mode the file is opened in to make sure that write
           is Allowed */
        if (((f->open_mode != FMOD_OUTPUT) && 
             (f->open_mode != FMOD_IO) && 
             (f->open_mode != FMOD_EXTEND)))
           RETURN_STATUS(48);

#ifdef DEBUG_FILEIO_RTS2
        fprintf(stderr, "debug : tcob_write_adv_lseq 0810.040 : f->reclen=%d\n", f->reclen); 
#endif     

        f->flags.with_advancing = 1;
        f->flags.adv_before = 0;

#ifdef DEBUG_FILEIO_RTS2
        fprintf(stderr, "debug : tcob_write_adv_lseq 0810.060 : opt=%d, lines=%d\n", opt, lines); 
#endif     
        if (opt > 0) {
           if (cnt_desc != NULL) 
              lines = tcob_get_index(cnt_desc, cnt_buf);
#ifdef DEBUG_FILEIO_RTS2
        fprintf(stderr, "debug : tcob_write_adv_lseq 0810.070 : opt=%d, lines=%d\n", opt, lines); 
#endif     
        }

#ifdef DEBUG_FILEIO_RTS2
        fprintf(stderr, "debug : tcob_write_adv_lseq 0810.080 : len=%d\n", len); 
#endif     
        if (reclen_desc != NULL) {
           tcob_move(reclen_desc, reclen_buf, &_generic_4binary, (char *)&len);     
        }
        else {
           len = f->reclen;
        }

#ifdef DEBUG_FILEIO_RTS2
        fprintf(stderr, "debug : tcob_write_adv_lseq 0810.120 : len=%d\n", len); 
#endif     

        /* remove trailling blanks */
        if (f->flags.file_printer == 0) {
           while (len > 0 && record[len-1] == ' ')
              len--;
        }
       
#ifdef DEBUG_FILEIO_RTS2
        fprintf(stderr, "debug : tcob_write_adv_lseq 0810.130 : len=%d\n", len); 
#endif     

        switch (opt)
        {
          /* 1: write record before advancing N lines */
          case 1:
#ifdef DEBUG_FILEIO_RTS2
            fprintf(stderr, "debug : tcob_write_adv_lseq 0810.210 : len=%d, lines=%d\n", len, lines); 
#endif     
            for (i=0; i<len; i++) 
              fputc(*(record + i), fdbp);
            fputc('\r', fdbp);
            for (i=lines; i>0; i--) 
                fputc('\n', fdbp);
            break;

          /* 2: write record after advancing N lines */
          case 2:
 #ifdef DEBUG_FILEIO_RTS2
            fprintf(stderr, "debug : tcob_write_adv_lseq 0810.220 : len=%d, lines=%d\n", len, lines); 
#endif     
           for (i=lines; i>0; i--) 
                fputc('\n', fdbp);
            for (i=0; i<len; i++) 
              fputc(*(record + i), fdbp);
            fputc('\r', fdbp);
            break;

          /* -1: write record before advancing page */
          case -1:
 #ifdef DEBUG_FILEIO_RTS2
            fprintf(stderr, "debug : tcob_write_adv_lseq 0810.230 : len=%d, lines=%d\n", len, lines); 
#endif     
            for (i=0; i<len; i++) 
              fputc(*(record + i), fdbp);
            fputc('\r', fdbp);
            fputc('\f', fdbp);
            break;

          /* -2: write record after advancing page */
          case -2:
#ifdef DEBUG_FILEIO_RTS2
            fprintf(stderr, "debug : tcob_write_adv_lseq 0810.240 : len=%d, lines=%d\n", len, lines); 
#endif     
            fputc('\f', fdbp);
            for (i=0; i<len; i++) 
              fputc(*(record + i), fdbp);
            fputc('\r', fdbp);
            break;

          /* write record before advancing 1 line */
          default:
#ifdef DEBUG_FILEIO_RTS2
            fprintf(stderr, "debug : tcob_write_adv_lseq 0810.260 : len=%d, lines=%d\n", len, lines); 
#endif     
            for (i=0; i<len; i++) 
              fputc(*(record + i), fdbp);
            fputc('\r', fdbp);
            fputc('\n', fdbp);
            break;
        }
        
        RETURN_STATUS(r);
}


//+---------------------------------------------------------------------------------+
// Sequential files use a constant length for fixed records and a unsigned int prefix                                                       
// for variable records.                                                      
// The variable format is defined by a unsigned int prefix plus the record length.                                                      
// The unsigned int prefix defines the actual length of that variable record.                                                      
//+---------------------------------------------------------------------------------+

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                        tcob_read_seq                                   |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_read_seq(struct file_desc *f, char *record, struct fld_desc *reclen_desc, char *reclen_buf) 
{
        int r=0, rpos=-1;

        f->flags.read_done = 0;
        
        /* Check to see if file is open. 
           If not return File Status 47 
           in accordance with the COBOL 85 Standard. */
        if (f->dbp == NULL) {
            if (f->flags.optional && f->flags.file_missing) 
               RETURN_STATUS(10);        
            RETURN_STATUS(47);
        }

        /* Check to see that the record length is valid */
        if (f->reclen == -1)
            RETURN_STATUS(99);
        
        /* Check the mode the file is opened in to make sure that read
           is Allowed */
        if (((f->open_mode != FMOD_INPUT) && (f->open_mode != FMOD_IO)))
           RETURN_STATUS(47);
        
        /* Check that we didn't already hit eof */
        if ((f->open_mode == FMOD_INPUT || f->open_mode == FMOD_IO) && f->flags.eof_hit)
           RETURN_STATUS(46);
        
#ifdef DEBUG_FILEIO_RTS1
        fprintf(stderr, "debug : tcob_read_seq 0720.060 : f->reclen=%d\n", f->reclen);
#endif                 

        r = tcob_read_rec(f, record, rpos, reclen_desc, reclen_buf); 

#ifdef DEBUG_FILEIO_RTS1
        fprintf(stderr, "debug : tcob_read_seq 0720.120 : r=%d\n", r);
#endif                 

        f->flags.read_done = 1;

        RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                        tcob_read_next_seq                              |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_read_next_seq(struct file_desc *f, char *record, struct fld_desc *reclen_desc, char *reclen_buf) 
{
  int r=0, rpos=-1;
  
  f->flags.read_done = 0;

  /* Check to see if file is open. 
     If not return File Status 47
     In accordance with the COBOL 85 Standard. */
  if (f->dbp == NULL) {
     if (f->flags.optional && f->flags.file_missing) 
        RETURN_STATUS(10);        
     RETURN_STATUS(47);
  }
  
  /* Check the mode the file is opened in to make sure that read is Allowed */
  if (((f->open_mode != FMOD_INPUT) && (f->open_mode != FMOD_IO)))
     RETURN_STATUS(47);
  
  /* Check that we didn't already hit eof */
  if ((f->open_mode == FMOD_INPUT || f->open_mode == FMOD_IO) && f->flags.eof_hit)
     RETURN_STATUS(46);
  
#ifdef DEBUG_FILEIO_RTS1
  fprintf(stderr, "debug : tcob_read_next_seq 0975.080 : f->reclen=%d\n", f->reclen);
#endif

  r = tcob_read_rec(f, record, rpos, reclen_desc, reclen_buf); 

#ifdef DEBUG_FILEIO_RTS1
  fprintf(stderr, "debug : tcob_read_next_seq 0975.160 : r=%d\n", r);
#endif

  f->flags.read_done = 1;

  RETURN_STATUS(r);

}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                        tcob_write_seq                                  |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_write_seq(struct file_desc *f, char *record, struct fld_desc *reclen_desc, char *reclen_buf) 
{
        int rpos=-1, r=0;
        
        f->flags.read_done = 0;
        
        /* Check to see if file is open. 
           If not return File Status 48
           In accordance with the COBOL 85 Standard. */
        if (f->dbp == NULL)
           RETURN_STATUS(48);
        
        /* Check to see that the record length is valid */
        if (f->reclen == -1) 
           RETURN_STATUS(48);

#ifdef DEBUG_FILEIO_RTS1
        fprintf(stderr, "debug : tcob_write_seq 0920.080 : f->reclen=%d\n", f->reclen);
#endif
       
        /* Check the mode the file is opened in to make sure that write is Allowed */
        if (((f->open_mode != FMOD_OUTPUT) && 
             (f->open_mode != FMOD_IO) && 
             (f->open_mode != FMOD_EXTEND)))
           RETURN_STATUS(48);

        r = tcob_write_rec(f, record, rpos, reclen_desc, reclen_buf); 
#ifdef DEBUG_FILEIO_RTS1
        fprintf(stderr, "debug : tcob_write_seq 0920.160 : r=%d\n", r);
#endif

        RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                        tcob_write_adv_seq                              |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_write_adv_seq(struct file_desc *f, char *record, int opt, 
                        struct fld_desc *cnt_desc, char *cnt_buf, 
                        struct fld_desc *reclen_desc, char *reclen_buf) 
{
        int result, fdbp, reclen, len, lines=0, r=0;

        /* Check to see if file is open. 
           If not return File Status 48
           In accordance with the COBOL 85 standard. */
        if (f->dbp == NULL)
           RETURN_STATUS(48);
        
        fdbp = (int)f->dbp;
        
        /* Check the mode the file is opened in to make sure that write
           is Allowed */
        if (((f->open_mode != FMOD_OUTPUT) && 
             (f->open_mode != FMOD_IO) && 
             (f->open_mode != FMOD_EXTEND)))
           RETURN_STATUS(48);

#ifdef DEBUG_FILEIO_RTS1
        fprintf(stderr, "debug : tcob_write_adv_seq 1090.040 : f->reclen=%d\n", f->reclen); 
#endif     

        if (opt > 0) {
           lines = tcob_get_index(cnt_desc, cnt_buf);
        }
        
        if (reclen_desc != NULL) {
           tcob_move(reclen_desc, reclen_buf, &_generic_4binary, (char *)&reclen);
           len = reclen;     
        }
        else {
           len = f->reclen;
        }

#ifdef DEBUG_FILEIO_RTS2
        fprintf(stderr, "debug : tcob_write_adv_seq 1090.120 : len=%d\n", len); 
#endif     

        // Check for buffer overflow
        reclen = len + lines + 2;
        if (reclen > (f->reclen * 2)) 
           RETURN_STATUS (99);
        
        f->flags.with_advancing = 1;
        f->flags.adv_before = 0;

        switch (opt)
        {
          /* 1: write before advancing N lines */
          case 1:
            memcpy(f->sbuf1, record, len);
            if (f->flags.file_printer == 1) {
               memset(f->sbuf1 + len, 0x0d, 1);
               memset(f->sbuf1 + len + 1, 0x0a, lines);
               reclen = len + lines + 1;
            }
            else {
               memset(f->sbuf1 + len, 0x0a, lines);
               reclen = len + lines;
            }
            break;

          /* 2: write after advancing N lines */
          case 2:
            if (f->flags.file_printer == 1) {
                memset(f->sbuf1, 0x0a, lines);
                memcpy(f->sbuf1 + lines, record, len);
                memset(f->sbuf1 + lines + len, 0x0d, 1);
                reclen = len + lines + 1;
            }
            else {
                memset(f->sbuf1, 0x0a, lines);
                memcpy(f->sbuf1 + lines, record, len);
                reclen = len + lines;
            }
            break;

          /* -1: before advancing page */
          case -1:
            if (f->flags.file_printer == 1) {
                memcpy(f->sbuf1, record, len);
                memset(f->sbuf1 + len, 0x0d, 1);            
                memset(f->sbuf1 + len + 1, 0x0c, 1);            
                reclen = len + 2;
            }
            else {
                memcpy(f->sbuf1, record, len);
                memset(f->sbuf1 + len, 0x0c, 1);            
                reclen = len + 1;
            }
            break;

          /* -2: after advancing page */
          case -2:
            if (f->flags.file_printer == 1) {
                memset(f->sbuf1, 0x0d, 1);            
                memset(f->sbuf1 + 1, 0x0c, 1);            
                memcpy(f->sbuf1 + 2, record, len);
                memset(f->sbuf1 + len + 2, 0x0d, 1);            
                reclen = len + 3;
            }
            else {
                memset(f->sbuf1 + 1, 0x0c, 1);            
                memcpy(f->sbuf1 + 1, record, len);
                reclen = len + 1;
            }
            break;

          /* only normal write (error?) */
          default:
            r = 99;
            break;
        }

        if (r == 0) {        
                  
            /* Variable length record */                  
            if (reclen_desc != NULL) {
               int sreclen;
               unsigned int ureclen;

               sreclen = sizeof(ureclen);
               ureclen = reclen;

               /* Write the record length */
               result = write(fdbp, &ureclen, sreclen);        
#ifdef DEBUG_FILEIO_RTS1
               fprintf(stderr, "debug : tcob_write_adv_seq 1090.240 : result=%d, sreclen=%d, ureclen=%d \n", result, sreclen, ureclen); 
#endif     
               if (result != sreclen)
                  RETURN_STATUS (99);
            }

            result = write(fdbp, f->sbuf1, reclen);
#ifdef DEBUG_FILEIO_RTS1
            fprintf(stderr, "debug : tcob_write_adv_seq 1090.280 : reclen=%d, result=%d\n", reclen, result);
#endif       
            if (result != reclen)
               RETURN_STATUS (30);
            if (result == -1)
               RETURN_STATUS (99);
       
        }

        RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                         tcob_rewrite_seq                               |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_rewrite_seq(struct file_desc *f, char *record, struct fld_desc *reclen_desc, char *reclen_buf) 
{
        int fdbp, r=0, srecpos, result;
        int read_done;
        
        read_done = f->flags.read_done;
        f->flags.read_done = 0;

        /* Check to see if file is open. 
           If not return File Status 49
           In accordance with the COBOL 85 standard. */
        if (f->dbp == NULL)
            RETURN_STATUS(49);
        fdbp = (int)f->dbp;
        
#ifdef DEBUG_FILEIO_RTS1
        fprintf(stderr, "debug : tcob_rewrite_seq 2458.020 : f->reclen=%d;\n", f->reclen); 
#endif     

        /* Check to see that the record length is valid */
        if (f->reclen == -1)
            RETURN_STATUS(99);
        
        /* Check the mode the file is opened in to make sure that rewrite
           is Allowed */
        if (f->open_mode != FMOD_IO)
           RETURN_STATUS(49);
        
        /* Check that we didn't already hit eof */
        if ((f->open_mode == FMOD_IO) && f->flags.eof_hit)
            RETURN_STATUS(43);
        
        /* Check that we had a previous successful read */
        if (!read_done)
            RETURN_STATUS(43);
        
#ifdef DEBUG_FILEIO_RTS1
        fprintf(stderr, "debug : tcob_rewrite_seq 2458.040 : f->reclen=%d;\n", f->reclen); 
#endif     

        /* Save current file location */
        result = lseek(fdbp, 0, SEEK_CUR);
#ifdef DEBUG_FILEIO_RTS1
        fprintf(stderr, "debug : tcob_rewrite_seq 2458.060 : result=%d;\n", result);
#endif
        if (result == -1)
           RETURN_STATUS(99);
        srecpos = result;

        // If variable length record check that new and old record length are equal 
        if (reclen_desc != NULL) {
           int reclen, sreclen;
           unsigned int ureclen;

           sreclen = sizeof(ureclen);

           // Get the new record  record length  
           tcob_move(reclen_desc, reclen_buf, &_generic_4binary, (char *)&reclen);     

           // Set file location to saved start position 
           result = lseek(fdbp, f->recloc, SEEK_SET);
#ifdef DEBUG_FILEIO_RTS1
           fprintf(stderr, "debug : tcob_rewrite_seq 2458.120 : result=%d\n", result);
#endif                 

           // Read record length 
           result = read(fdbp, (char *)&ureclen, sreclen);
#ifdef DEBUG_FILEIO_RTS1
           fprintf(stderr, "debug : tcob_rewrite_seq 2458.160 : sreclen=%d, result=%d\n", sreclen, result);
#endif                 
           // On error try to reset the old file location
           if (result == 0) {
              result = lseek(fdbp, srecpos, SEEK_SET);
              RETURN_STATUS(10);
           }

#ifdef DEBUG_FILEIO_RTS1
           fprintf(stderr, "debug : tcob_rewrite_seq 2458.180 : reclen=%d, ureclen=%d, f->reclen=%d\n", reclen, ureclen, f->reclen);
#endif                 
           // On error try to reset the old file location
           if (reclen != ureclen) {
              result = lseek(fdbp, srecpos, SEEK_SET);
              RETURN_STATUS(30);
           }
        }
        
        /* Write the new record */
        if (r == 0)
           r = tcob_write_rec(f, record, f->recloc, reclen_desc, reclen_buf); 
#ifdef DEBUG_FILEIO_RTS1
        fprintf(stderr, "debug : tcob_rewrite_seq 2458.240 : r=%d;\n", r);
#endif

        RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                         tcob_read_rec                                  |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_read_rec(struct file_desc *f, char *record, long int recpos, struct fld_desc *reclen_desc, char *reclen_buf) 
{
  int r=0, result, fdbp, crecpos;

  fdbp = (int)f->dbp;

  if (recpos > -1)  {
     struct stat stbuf;
     
     result = fstat(fdbp, &stbuf);
     if (result == -1) 
        return (99);

     if (stbuf.st_size < recpos) 
        return (99);
     
     result = lseek(fdbp, recpos, SEEK_SET);
     if (result == -1) 
        return (23);
  }
  
  crecpos = lseek(fdbp, 0, SEEK_CUR);

  // Variable Length Sequential File 
  if (reclen_desc != NULL) {
     int reclen, sreclen;
     unsigned int ureclen;
         
     sreclen = sizeof(ureclen);

     // Read record length 
     result = read(fdbp, (char *)&ureclen, sreclen);
#ifdef DEBUG_FILEIO_RTS1
     fprintf(stderr, "debug : tcob_read_rec 2440.120 : result=%d, sreclen=%d\n", result, sreclen);
#endif                 
     if (result == 0)
        return (10);

     reclen = ureclen;
     tcob_move(&_generic_4binary, (char *)&reclen, reclen_desc, reclen_buf);     

     // Read data using the read-in record length 
     memset(record, ' ', f->reclen);
     result = read(fdbp, record, reclen);
#ifdef DEBUG_FILEIO_RTS1
     fprintf(stderr, "debug : tcob_read_rec 2440.240 : result=%d, reclen=%d\n", result, reclen);
     {
      int i, ilen=reclen;
      fprintf(stderr, "debug : tcob_read_rec 2440.145 : record=");
      for (i=0; i<ilen; i++) {      
         fprintf(stderr, "%c", *(record + i));
      }
      fprintf(stderr, ";\n");
     }
#endif                 
     if (result > f->reclen)
         return (99);
     if (result == 0)
         return (10);
     if (result != reclen)
         return (30);
  }
  else {     
      result = read(fdbp, record, f->reclen);
#ifdef DEBUG_FILEIO_RTS1
     fprintf(stderr, "debug : tcob_read_rec 2440.260 : result=%d, f->reclen=%d\n", result, f->reclen);
#endif                 
      if (result == 0)
         return (10);
      if (result != f->reclen)
         return (30);
  }

  /* save the record start position */
  f->recloc = crecpos;  

  return r;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                        tcob_write_rec                                  |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_write_rec(struct file_desc *f, char *record, int recpos, struct fld_desc *reclen_desc, char *reclen_buf) 
{
  int r=0, result, fdbp;

  fdbp = (int)f->dbp;

  if (recpos > -1)  {
     struct stat stbuf;
     
     result = fstat(fdbp, &stbuf);
     if (result == -1) 
        return (99);

     if (stbuf.st_size < recpos) 
        return (99);
     
     result = lseek(fdbp, recpos, SEEK_SET);
#ifdef DEBUG_FILEIO_RTS1
  fprintf(stderr, "debug : tcob_write_rec 2490.060 : result=%d, recpos=%d;\n", result, recpos); 
#endif     
     if (result == -1) 
        return (99);
  }

#ifdef DEBUG_FILEIO_RTS1
  fprintf(stderr, "debug : tcob_write_rec 2490.080 : f->recloc=%d\n", f->recloc); 
#endif     
  
  /* Variable length record */                  
  if (reclen_desc != NULL) {
     int reclen, sreclen;
     unsigned int ureclen;

     sreclen = sizeof(ureclen);
     /* Convert record to int type */
     tcob_move(reclen_desc, reclen_buf, &_generic_4binary, (char *)&reclen);     
     ureclen = reclen;

#ifdef DEBUG_FILEIO_RTS1
     fprintf(stderr, "debug : tcob_write_rec 2490.120 : ureclen=%d, reclen=%d\n", ureclen, reclen); 
     {
      int i, ilen=reclen;
      fprintf(stderr, "debug : tcob_write_rec 2490.125 : record=");
      for (i=0; i<ilen; i++) {      
         fprintf(stderr, "%c", *(record + i));
      }
      fprintf(stderr, ";\n");
     }
#endif     
        
     /* Write the record length */
     result = write(fdbp, &ureclen, sreclen);        
#ifdef DEBUG_FILEIO_RTS1
     fprintf(stderr, "debug : tcob_write_rec 2490.180 : result=%d, sreclen=%d\n", result, sreclen); 
#endif     
     if (result > f->reclen)
        return (99);
     if (result == 0)
        return (10);

     /* Write the record  */
     result = write(fdbp, record, reclen);        
#ifdef DEBUG_FILEIO_RTS1
     fprintf(stderr, "debug : tcob_write_rec 2490.240 : reclen=%d, result=%d\n", reclen, result);
#endif       
     if (result > f->reclen)
        return (99);
     if (result == 0)
        return (10);
     if (result != reclen)
        return (30);
  } 
  else {
     result = write(fdbp, record, f->reclen);
#ifdef DEBUG_FILEIO_RTS1
     fprintf(stderr, "debug : tcob_write_rec 2490.280 : f->reclen=%d, result=%d\n", f->reclen, result);
     {
      int i, ilen=f->reclen;
      fprintf(stderr, "debug : tcob_write_rec 2490.285 : record(%d)=", ilen);
      for (i=0; i<ilen; i++) {      
          fprintf(stderr, "%c", *(record + i));
      }
      fprintf(stderr, ";\n");
     }
#endif
     if (result == 0)
        return (99);
     if (result != f->reclen)
        return (30);
  }

  return r;
}

//+---------------------------------------------------------------------------------+
// Relative files use a constant record length for both variable and fixed record.                                                      
// The format is defined by a unsigned int32 prefix plus the max record length.                                                      
// The unsigned int32 prefix defines the real length of that record.                                                      
// A real record length of zero indicates that the record is not defined.                                                      
//+---------------------------------------------------------------------------------+
/*------------------------------------------------------------------------*\
 |                                                                        |
 |                        tcob_read_rel                                   |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_read_rel(struct file_desc *f, char *record, 
                  struct fld_desc *cnt_desc, char *cnt_buf, 
                  struct fld_desc *reclen_desc, char *reclen_buf) 
{
        int result, fdbp, recpos = -1, sreclen, r=0;
        unsigned int ureclen;
        unsigned int recno=0;

        fdbp = (int)f->dbp;
        f->flags.read_done = 0;
        sreclen = sizeof(ureclen);

        /* Check to see if file is open. 
           If not return File Status 47
           in accordance with the COBOL 85 standard. */
        if (f->dbp == NULL) {
            if (f->flags.optional && f->flags.file_missing) 
               RETURN_STATUS(10);        
            RETURN_STATUS(47);
        }
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_read_rel : 1170.040 : f->reclen=%d\n", f->reclen);
#endif

        /* Check to see that the record length is valid */
        if (f->reclen == -1)
            RETURN_STATUS(99);
        
        /* 
         Check the mode the file is opened in to make sure that read
         is Allowed 
        */
        if (((f->open_mode != FMOD_INPUT) && (f->open_mode != FMOD_IO)))
           RETURN_STATUS(47);
        
        /* Check that we didn't already hit eof */
        if ((f->open_mode == FMOD_INPUT || f->open_mode == FMOD_IO) && f->flags.eof_hit)
           RETURN_STATUS(46);
        
        if (cnt_desc != NULL) 
            recno = tcob_get_index(cnt_desc, cnt_buf);
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_read_rel : 1170.080 : recno=%d\n", recno); 
#endif                                       

        if ((recno < 1) && (f->access_mode != ACC_SEQUENTIAL)) 
           RETURN_STATUS(23);

        /* If there is a start record outstanding use it to fulfill the read */ 
        if (f->flags.start_record == 1) {
           recpos = f->recloc;
           f->flags.start_record = 0; 
           result = lseek(fdbp, recpos, SEEK_SET);
#ifdef DEBUG_FILEIO_RTS3
           fprintf(stderr, "debug : tcob_read_rel : 1170.100 : result=%d\n", result);
#endif                 
           if (result == -1)
             RETURN_STATUS(99);
        }
        else {
//        if (recno < 1) && (f->access_mode == SEQUENTIAL))
           /* If record number given set the file location */
           if (recno > 0) {
              /* Use the max record length and record number to set the file location */
              recpos = (recno - 1) * (f->reclen + sizeof(int));
              result = lseek(fdbp, recpos, SEEK_SET);
#ifdef DEBUG_FILEIO_RTS3
              fprintf(stderr, "debug : tcob_read_rel : 1170.110 : result=%d\n", result);
#endif                 
              if(result == -1)
                RETURN_STATUS(23);
           }
        }

        // Fixed and variable length relative files         
        // Read record length 
        result = read(fdbp, (char *)&ureclen, sreclen);
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_read_rel : 1170.120 : result=%d, sreclen=%d, ureclen=%d\n", result, sreclen, ureclen);
#endif                 
        if (result == 0)
           RETURN_STATUS(23);
        if (ureclen > f->reclen)
           RETURN_STATUS(99);
        // Check if record exists (f->reclen >= ureclen > 0)
        if (ureclen == 0) {
           recpos = -1 * sreclen;
           result = lseek(fdbp, recpos, SEEK_CUR);
           if (result == 0)
              RETURN_STATUS(99);           
           RETURN_STATUS(23);
        }

        // Read data using the full (max) record length 
        result = read(fdbp, f->sbuf1, f->reclen);

#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_read_rel : 1170.240 : result=%d, ureclen=%d, f->reclen=%d\n", result, ureclen, f->reclen);
#endif                 

        if (result != f->reclen)
            RETURN_STATUS(30);
        if (result == 0)
            RETURN_STATUS(10);

#ifdef DEBUG_FILEIO_RTS3
        {
         int i, ilen=ureclen;
         fprintf(stderr, "debug : tcob_read_rel : 1170.245 : sbuf1(%d)=", ilen);
         for (i=0; i<ilen; i++) {      
            fprintf(stderr, "%c", *(f->sbuf1 + i));
         }
         fprintf(stderr, ";\n");
        }
#endif                 

        // If variable length file move record length data to decriptor 
        if (reclen_desc != NULL) {
           int reclen = ureclen;
           tcob_move(&_generic_4binary, (char *)&reclen, reclen_desc, reclen_buf);     
        }
//        memset(record, ' ', f->reclen);
        memcpy(record, f->sbuf1, ureclen);
        f->flags.read_done = 1;

        RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                        tcob_read_next_rel                              |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_read_next_rel(struct file_desc *f, char *record, struct fld_desc *reclen_desc, char *reclen_buf) 
{
  int result, fdbp, reclen, sreclen, recpos=-1, r=0;
  unsigned int ureclen;
  
  sreclen = sizeof(ureclen);
  
  /* Check to see if file is open. 
     If FALSE return file-status 47. 
     In accordance with the COBOL 85 standard. */
  if (f->dbp == NULL) {
     if (f->flags.optional && f->flags.file_missing) 
        RETURN_STATUS(10);        
     RETURN_STATUS(47);
  }
  fdbp = (int)f->dbp;
  
  /* Check the mode the file is opened in to make sure that read is Allowed */
  if (((f->open_mode != FMOD_INPUT) && (f->open_mode != FMOD_IO)))
     RETURN_STATUS(47);
  
  /* Check that we didn't already hit eof */
  if ((f->open_mode == FMOD_INPUT || f->open_mode == FMOD_IO) && f->flags.eof_hit)
     RETURN_STATUS(46);
  
  reclen = f->reclen + sizeof(int);

  /* If there is a start record outstanding use it to fulfill the read */
  if (f->flags.start_record == 1) {
     recpos = f->recloc;
     f->flags.start_record = 0; 
     result = lseek(fdbp, recpos, SEEK_SET);
#ifdef DEBUG_FILEIO_RTS3
     fprintf(stderr, "debug : tcob_read_next_rel : 1700.120 : result=%d\n", result);
#endif                 
     if (result == -1)
        RETURN_STATUS(99);
     // Read record length 
     result = read(fdbp, (char *)&ureclen, sreclen); 
#ifdef DEBUG_FILEIO_RTS3
     fprintf(stderr, "debug : tcob_read_next_rel : 1700.130 : result=%d, sreclen=%d, ureclen=%d\n", result, sreclen, ureclen);
#endif                 
     if (result != sreclen)
        RETURN_STATUS(99);           
     // Check if record exists (f->reclen >= ureclen > 0)
     if (ureclen == 0) 
        RETURN_STATUS(23);
  }
  else {
     int rpos, sw=0;
        
     rpos = lseek(fdbp, 0, SEEK_CUR);
#ifdef DEBUG_FILEIO_RTS3
     fprintf(stderr, "debug : tcob_read_next_rel : 1700.220 : rpos=%d\n", rpos);
#endif                 
     while (sw == 0) {
       // Read record length 
       result = read(fdbp, (char *)&ureclen, sreclen); 
#ifdef DEBUG_FILEIO_RTS3
       fprintf(stderr, "debug : tcob_read_next_rel : 1700.240 : result=%d, sreclen=%d, ureclen=%d\n", result, sreclen, ureclen);
#endif                 
       if (result == 0)
          RETURN_STATUS(10);           
       if (result != sreclen)
          RETURN_STATUS(99);           
       // Check if record exists (f->reclen >= ureclen > 0)
       if (ureclen != 0) {
          sw = 1;
       }
       else {
          rpos = lseek(fdbp, f->reclen, SEEK_CUR);
#ifdef DEBUG_FILEIO_RTS3
          fprintf(stderr, "debug : tcob_read_next_rel : 1700.260 : rpos=%d\n", rpos);
#endif                 
       }
     }                
  }
  
  // Read data using the full (max) record length 
  result = read(fdbp, f->sbuf1, f->reclen);
#ifdef DEBUG_FILEIO_RTS3
  fprintf(stderr, "debug : tcob_read_next_rel : 1700.320 : result=%d, ureclen=%d, f->reclen=%d\n", result, ureclen, f->reclen);
#endif                 

  if (result != f->reclen)
      RETURN_STATUS(30);
  if (result == 0)
      RETURN_STATUS(23);

#ifdef DEBUG_FILEIO_RTS3
  {
   int i, ilen=ureclen;
   fprintf(stderr, "debug : tcob_read_next_rel : 1700.330 : sbuf1(%d)=", ilen);
   for (i=0; i<ilen; i++) {      
      fprintf(stderr, "%c", *(f->sbuf1 + i));
   }
   fprintf(stderr, ";\n");
  }
#endif                 

  // If variable length file move record length data to decriptor 
  if (reclen_desc != NULL) {
     reclen = ureclen;
     tcob_move(&_generic_4binary, (char *)&reclen, reclen_desc, reclen_buf);     
  }

//  memset(record, ' ', f->reclen);
  memcpy(record, f->sbuf1, ureclen);
  f->flags.read_done = 1;

  RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                        tcob_write_rel                                  |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_write_rel(struct file_desc *f, char *record,                   
                   struct fld_desc *cnt_desc, char *cnt_buf, 
                   struct fld_desc *reclen_desc, char *reclen_buf) 
{
        int result=0, recpos, fdbp, sreclen, r=0;
        unsigned int ureclen;
        unsigned int recno=0;
        
        f->flags.read_done = 0;
        sreclen = sizeof(ureclen);
        
        /* Check to see if file is open. 
           If not return File Status 48
           In accordance with the COBOL 85 standard. */
        if (f->dbp == NULL)
           RETURN_STATUS(48);
        fdbp = (int)f->dbp;
        
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_write_rel : 1810.040 : f->reclen=%d\n", f->reclen);
#endif
        /* Check to see that the record length is valid */
        if (f->reclen == -1) 
           RETURN_STATUS(99);
      
        /* Check the mode the file is opened in to make sure that write is Allowed */
        if (((f->open_mode != FMOD_OUTPUT) && 
             (f->open_mode != FMOD_IO) && 
             (f->open_mode != FMOD_EXTEND)))
           RETURN_STATUS(48);

        if (cnt_desc != NULL) 
            recno = tcob_get_index(cnt_desc, cnt_buf);
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_write_rel : 1810.080 : recno=%d\n", recno); 
#endif                                       

//        if (recno < 1) && (f->access_mode == SEQUENTIAL))
        if (f->access_mode != ACC_SEQUENTIAL) {
           if (recno < 1)
              RETURN_STATUS(21);
          /* Use the max record length and record number to set the file location */
          recpos = (recno - 1) * (f->reclen + sizeof(int));
          result = lseek(fdbp, recpos, SEEK_SET);
#ifdef DEBUG_FILEIO_RTS3
          fprintf(stderr, "debug : tcob_write_rel : 1810.090 : recpos=%d, result=%d;\n", recpos, result); 
#endif                                       
          if (result == -1) 
            RETURN_STATUS(21);
        }
        
        // Read record length 
        result = read(fdbp, (char *)&ureclen, sreclen);
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_write_rel : 1810.120 : result=%d, sreclen=%d, ureclen=%d;\n", result, sreclen, ureclen); 
#endif                                       
        if (result == -1)
           RETURN_STATUS(99);
        if (result > 0) {
           // Re-set file position
           recpos = -1 * sreclen;
           result = lseek(fdbp, recpos, SEEK_CUR);
#ifdef DEBUG_FILEIO_RTS3
           fprintf(stderr, "debug : tcob_write_rel : 1810.130 : recpos=%d, result=%d;\n", recpos, result); 
#endif                                       
//           if (result == 0)
           if (result < 0)
              RETURN_STATUS(99);           
           // Key exists 
           if (ureclen > 0) 
              RETURN_STATUS(22);
        }
        
        // If variable record length 
        if (reclen_desc != NULL) {
          int reclen;
          /* Convert record to unsigned int type */
          tcob_move(reclen_desc, reclen_buf, &_generic_4binary, (char *)&reclen);     
          ureclen = reclen;
        }
        else {
          ureclen = f->reclen;        
        }
       
        memset(f->sbuf1, ' ', f->reclen + sreclen);
        memcpy(f->sbuf1, &ureclen, sreclen);
        memcpy(f->sbuf1 + sreclen, record, ureclen);

#ifdef DEBUG_FILEIO_RTS3
        {
         fprintf(stderr, "debug : tcob_write_rel : 1810.160 : ureclen=%d, f->reclen=%d\n", ureclen, f->reclen); 
         int i, ilen=ureclen;
         fprintf(stderr, "debug : tcob_write_rel : 1810.165 : record(%d)=", ilen);
         for (i=0; i<ilen; i++) {      
            fprintf(stderr, "%c", *(record + i));
         }
         fprintf(stderr, ";\n");
        }
#endif     

        result = write(fdbp, f->sbuf1, f->reclen + sreclen);        
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_write_rel : 1810.240 : result=%d, f->reclen=%d, sreclen=%d\n", result, f->reclen, sreclen);
#endif       

        if (result == 0)
           return (10);
        if (result != (sreclen + f->reclen))
           return (99);

        RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                        tcob_rewrite_rel                                |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_rewrite_rel(struct file_desc *f, char *record, 
                     struct fld_desc *cnt_desc, char *cnt_buf, 
                     struct fld_desc *reclen_desc, char *reclen_buf) 
{
        int result, fdbp, sreclen, reclen, recpos, r=0;
        unsigned int ureclen;
        unsigned int recno=0;
        
        /* Check to see if file is open. 
           If not return a file-status 49.
           In accordance with the COBOL 85 standard. */
        if (f->dbp == NULL)
            RETURN_STATUS(49);
        fdbp = (int)f->dbp;
        sreclen = sizeof(ureclen);

        /* Check to see that the record length is valid */
        if (f->reclen == -1)
            RETURN_STATUS(99);
        
        /* Check the mode the file is opened in to make sure that rewrite is Allowed */
        if (f->open_mode != FMOD_IO)
           RETURN_STATUS(49);
                
        /* Check that we had a previous successful read */
//        if (!read_done)
//            RETURN_STATUS(43);
        
        if (cnt_desc != NULL) 
            recno = tcob_get_index(cnt_desc, cnt_buf);
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_rewrite_rel : 2610.080 : recno=%d\n", recno); 
#endif                                       
        if ((recno < 1) && 
            (f->access_mode != ACC_SEQUENTIAL))
           RETURN_STATUS(23);

        /* If record number is NOT given INVALID KEY */
        if (recno < 1)
           RETURN_STATUS(21);
        
        /* Use the max record length and record number to set the file location */
        recpos = (recno - 1) * (f->reclen + sizeof(int));
        result = lseek(fdbp, recpos, SEEK_SET);
 #ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_rewrite_rel : 2610.090 : result=%d\n", result); 
#endif                                       
        if (result == -1)
            RETURN_STATUS(99);
        
        
        /* All relative files are variable-fixed record length */
        // Read record length 
        result = read(fdbp, (char *)&ureclen, sreclen);
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_rewrite_rel : 2610.100 : result=%d, sreclen=%d, ureclen=%d\n", result, sreclen, ureclen);
#endif                 
        if (result == 0)
           RETURN_STATUS(10);
        if (ureclen > f->reclen)
           RETURN_STATUS(99);
        // Check if record exists (f->reclen >= ureclen > 0)
        if (ureclen == 0) {
           recpos = -1 * sreclen;
           result = lseek(fdbp, recpos, SEEK_SET);
           if (result == 0)
              RETURN_STATUS(99);           
           RETURN_STATUS(23);
        }

        // Get the variable record description length 
        if (reclen_desc != NULL) {
           tcob_move(reclen_desc, reclen_buf, &_generic_4binary, (char *)&reclen);     
#ifdef DEBUG_FILEIO_RTS3
           fprintf(stderr, "debug : tcob_rewrite_rel : 2610.120 : f->reclen=%d, reclen=%d, ureclen=%d\n", f->reclen, reclen, ureclen); 
#endif     
        }
        else {
           reclen = f->reclen;
        }

        if (reclen != ureclen) {
            recpos = sreclen * -1;            
            result = lseek(fdbp, recpos, SEEK_CUR);
            RETURN_STATUS(99);
        }

        memset(f->sbuf1, ' ', f->reclen);
        memcpy(f->sbuf1, record, ureclen);

        result = write(fdbp, f->sbuf1, f->reclen);        

#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_rewrite_rel : 2610.180 : result=%d\n", result);
#endif     

        if (result == 0)
           RETURN_STATUS(10);
        if (result != f->reclen)
           RETURN_STATUS(99);
             
#ifdef DEBUG_FILEIO_RTS3
        {
         int i, ilen=ureclen;
         fprintf(stderr, "debug : tcob_rewrite_rel : 2610.190 : sbuf1(%d)=", ilen);
         for (i=0; i<ilen; i++) {      
            fprintf(stderr, "%c", *(f->sbuf1 + i));
         }
        }
        fprintf(stderr, ";\n");
#endif     

        RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                      tcob_delete_rel                                   |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_delete_rel(struct file_desc *f, char *record, struct fld_desc *cnt_desc, char *cnt_buf) 
{
        int result, fdbp, sreclen, recpos, r=0;
        unsigned int ureclen;
        unsigned int recno=0;
        
        /* Check to see if file is open, 
           if not return a file-status 49
           In accordance with the COBOL 85 standard.  */
        if (f->dbp == NULL)
           RETURN_STATUS(49);

        fdbp = (int)f->dbp;
        sreclen = sizeof(ureclen);
        
        // Check to see that the record length is valid 
        if (f->reclen == -1)
           RETURN_STATUS(99);
         
        //  Verify that the file must be open in I-O mode.
        if (f->open_mode != FMOD_IO)
           RETURN_STATUS(49);
        
        if (cnt_desc != NULL) 
            recno = tcob_get_index(cnt_desc, cnt_buf);
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_delete_rel : 2045.080 : recno=%d\n", recno); 
#endif                                       
        if ((recno < 1) && (f->access_mode != ACC_SEQUENTIAL))  
           RETURN_STATUS(23);

        if (f->access_mode != ACC_SEQUENTIAL) {
           if (recno < 1)
              RETURN_STATUS(21);
          /* Use the max record length and record number to set the file location */
          recpos = (recno - 1) * (f->reclen + sizeof(int));
          result = lseek(fdbp, recpos, SEEK_SET);
#ifdef DEBUG_FILEIO_RTS3
          fprintf(stderr, "debug : tcob_delete_rel : 2045.090 : recpos=%d, result=%d;\n", recpos, result); 
#endif                                       
          if (result == -1) 
            RETURN_STATUS(21);
        }
        else {
          if (f->flags.read_done == 0)
            RETURN_STATUS(21);
          recpos = -1 * (f->reclen + sizeof(int));
          result = lseek(fdbp, recpos, SEEK_CUR);
          if (result == -1) 
            RETURN_STATUS(21);
        }

        // Read record length 
        result = read(fdbp, (char *)&ureclen, sreclen);
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_delete_rel : 2045.100 : result=%d, sreclen=%d, ureclen=%d\n", result, sreclen, ureclen);
#endif                 
        if (result == 0)
           RETURN_STATUS(10);
        if (ureclen > f->reclen)
           RETURN_STATUS(99);
        // Check if record exists (f->reclen >= ureclen > 0)
        if (ureclen == 0) 
           RETURN_STATUS(23);
        
        recpos = -1 * sreclen;
//        result = lseek(fdbp, recpos, SEEK_SET);
        result = lseek(fdbp, recpos, SEEK_CUR);
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_delete_rel : 2045.120 : recpos=%d, result=%d;\n", recpos, result); 
#endif                 
        if (result == 0)
           RETURN_STATUS(99);           

        ureclen = 0;
        memset(f->sbuf1, ' ', f->reclen + sreclen);
        memcpy(f->sbuf1, &ureclen, sreclen);

        result = write(fdbp, f->sbuf1, f->reclen + sreclen);        
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_delete_rel : 2045.180 : result=%d, f->reclen=%d, sreclen=%d\n", result, f->reclen, sreclen);
#endif       

        if (result == 0)
           return (10);
        if (result != (sreclen + f->reclen))
           return (99);

        RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_start_rel                                |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_start_rel(struct file_desc *f, char *record, int cond, struct fld_desc *cnt_desc, char *cnt_buf) 
{
        int result, fdbp, reclen, sreclen, recpos=-1, r=0;
        unsigned int ureclen;
        struct stat stbuf;
        unsigned int recno=0;
                
        sreclen = sizeof(ureclen);
        f->flags.eof_hit = 0;  /* additon AL : reset eof if already positioned */
        f->flags.read_done = 0;
        f->flags.start_record = 0;
        
        /* Check to see if file is open. 
           If not return File Status 47
           In accordance with the COBOL 85 standard. */
        if (f->dbp == NULL)
           RETURN_STATUS(47);

        fdbp = (int)f->dbp;
       
        /* Check to see that the record length is valid */
        if (f->reclen == -1)
           RETURN_STATUS(99);
        
        /* 
         Check the mode the file is opened in to make sure 
         that start is allowed 
        */
        if (((f->open_mode != FMOD_INPUT) && (f->open_mode != FMOD_IO)))
           RETURN_STATUS(47);
        
        if (cnt_desc != NULL) 
            recno = tcob_get_index(cnt_desc, cnt_buf);
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_start_rel : 1495.040: recno=%d\n", recno); 
#endif                                       

        reclen = f->reclen + sizeof(int);

        result = fstat(fdbp, &stbuf);
        if (result == -1) 
           RETURN_STATUS(99);     
#ifdef DEBUG_FILEIO_RTS3
        fprintf(stderr, "debug : tcob_start_rel : 1495.080 : result=%d, stbuf.st_size=%ld\n", result, stbuf.st_size);
#endif                 
        
        switch(cond) {
          /* case 1:  Equal to */
          case TCB_CONDITIONAL_EQUAL: /* Equal to */
               {
                int rno;

                rno = (recno - 1);
                recpos = rno * reclen;
                if (stbuf.st_size < recpos) 
                   RETURN_STATUS(23);
                result = lseek(fdbp, recpos, SEEK_SET);
                if (result != recpos) 
                   RETURN_STATUS(23);
                // Read record length 
                result = read(fdbp, (char *)&ureclen, sreclen);
#ifdef DEBUG_FILEIO_RTS3
                fprintf(stderr, "debug : tcob_start_rel : 1495.120 : result=%d, sreclen=%d, ureclen=%d\n", result, sreclen, ureclen);
#endif                 
                recpos = -1 * sreclen;
                recpos = lseek(fdbp, result, SEEK_CUR);
                // Check if record exists (f->reclen >= ureclen > 0)
                if (ureclen == 0) {
                   if (recpos == -1)
                      RETURN_STATUS(99);           
                   RETURN_STATUS(23);
                }
                f->recloc = rno * reclen;
                f->flags.start_record = 1;
               }
               break;

          /* case 4: Greater Than */
          case TCB_CONDITIONAL_GT: 
               {
                int rpos, rno, sw=0;
                
                rno = recno;
                recpos = rno * reclen;
                while (sw == 0) {
                  rpos = rno * reclen;
                  if (stbuf.st_size < rpos) 
                     RETURN_STATUS(23);
                  result = lseek(fdbp, rpos, SEEK_SET);
#ifdef DEBUG_FILEIO_RTS3
                  fprintf(stderr, "debug : tcob_start_rel : 1495.420 : result=%d, reclen=%d, rpos=%d\n", result, reclen, rpos);
#endif                 
                  if (result != rpos) 
                     RETURN_STATUS(23);
                  // Read record length 
                  result = read(fdbp, (char *)&ureclen, sreclen);
#ifdef DEBUG_FILEIO_RTS3
                  fprintf(stderr, "debug : tcob_start_rel : 1495.460 : result=%d, sreclen=%d, ureclen=%d\n", result, sreclen, ureclen);
#endif                 
                  if (result != sreclen)
                     RETURN_STATUS(99);           
                  // Check if record exists (f->reclen >= ureclen > 0)
                  if (ureclen != 0) {
                     f->recloc = rno * reclen;
                     f->flags.start_record = 1;
                     sw = 0;
                     break;
                  }
                  rno++;
                }                
               }                
               break;

          /* case 5: Greater than or Equal to */
          case TCB_CONDITIONAL_GTEQ:
               {
                int rpos, rno, sw=0;
                
                rno = recno - 1;
                recpos = rno * reclen;
                while (sw == 0) {
                  rpos = rno * reclen;
                  if (stbuf.st_size < rpos) 
                     RETURN_STATUS(23);
                  result = lseek(fdbp, rpos, SEEK_SET);
                  if (result != rpos) 
                     RETURN_STATUS(23);
                  // Read record length 
                  result = read(fdbp, (char *)&ureclen, sreclen);
#ifdef DEBUG_FILEIO_RTS3
                  fprintf(stderr, "debug : tcob_start_rel : 1495.520 : result=%d, sreclen=%d, ureclen=%d\n", result, sreclen, ureclen);
#endif                 
                  if (result != sreclen)
                     RETURN_STATUS(99);           
                  // Check if record exists (f->reclen >= ureclen > 0)
                  if (ureclen != 0) {
                     f->recloc = rno * reclen;
                     f->flags.start_record = 1;
                     sw = 0;
                     break;
                  }
                  rno++;
                }                
               }                
               break;

          
          case TCB_CONDITIONAL_LT:   /* case 2: Less Than */
          case TCB_CONDITIONAL_LTEQ: /* case 3: Less than or equal to */
          default:
               r = 99;
               break;
        }

        RETURN_STATUS(r);
}

/* EOF fileio.c */
