/*
 * Copyright (C) 2007        Rildo Pragana, David Essex.
 * Copyright (C) 1999-2004,  Rildo Pragana, Ferran Pegueroles, Jim Noeth,
 *                           Bernard Giroud, David Essex, Andrew Cameron. 
 * Copyright (C) 1993, 1991  Rildo Pragana.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1,
 * or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; see the file COPYING.LIB.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA 02111-1307 USA
  
 * 
 * TinyCOBOL Run Time Library -- BDB Indexed file handling functions
 * 

*/

#include "htcoblib.h"
#include "fileio.h" 
#include "fileiodb.h"

//#define DEBUG_FILEDBIO_RTS0    1   /* Open and close statements */ 
//#define DEBUG_FILEDBIO_RTS1    1   /* Read [ Next ] statements  */
//#define DEBUG_FILEDBIO_RTS2    1   /* Write statement           */
//#define DEBUG_FILEDBIO_RTS4    1   /* Start statement           */
//#define DEBUG_FILEDBIO_RTS5    1   /* Delete statement          */
//#define DEBUG_FILEDBIO_RTS6    1   /* ReWrite statement         */
 
extern void save_status (struct file_desc *f, int status);
extern int tcob_save_status(char *status, int rt);

#define db_open dbopen

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_open_idx                                  |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_open_idx(struct file_desc *f, char *filename, int mode, int oflags, int sflags) 
{
  int r=0, len, fdbp;
  char *fn;
  DB *fdbpx;  
  DBTYPE type;
  BTREEINFO xkey;

  type=DB_BTREE;
  
  xkey.flags = 0;
  xkey.cachesize = 0;
  xkey.maxkeypage = 0;
  xkey.minkeypage = 0;
  xkey.psize = 0;
  xkey.compare = NULL;
  xkey.prefix = NULL;
  xkey.lorder = 0;
         
#ifdef DEBUG_FILEDBIO_RTS0
  fprintf(stderr, "debug : tcob_open_idx : 0080.020 : filename=%s, oflags=%d, sflags=%d;\n", filename, oflags, sflags); 
#endif                                       
  
  len = strlen(filename);
  fn = (char *)malloc((sizeof(char) * len) + 10);

  /* FIXME: file suffix should be user defined */
  /* Open sequential data file */
  strncpy(fn, filename, len + 1);
  strncat(fn, ".dat", 5);
  fdbp = open(fn, oflags, sflags);
//  free(dbpname);
  if (fdbp == -1) {
     free(fn);
     f->dbp=0;
//     r = 99;
     r = 35;
     return r;
  }
  
  /* FIXME: file suffix should be user defined */
  /* Open BDB index file */
  strncpy(fn, filename, len + 1);
  strncat(fn, ".dbx", 5);  
  fdbpx = db_open(fn, oflags, sflags, type, &xkey);
  free(fn);
  if (fdbpx == NULL) {
     close(fdbp);
     r = 91;
     if (errno == EINVAL) {
        r = 37;
     }
     if (errno == ENOENT) {
        r = 35;
     }
     f->dbpx = NULL; 
     return r;
  }

  f->recloc = -1;
  f->dbp = (void *)fdbp;
  f->dbpx = (void *)fdbpx;

  /* save mode to check later (read,write,start,...)  */
  f->open_mode = mode;
  /* save the file in the list of open files for the proces */
  tcob_add_file_list(f);

  /* File IO record buffer */ 
  f->sbuf1 = (char *)malloc((sizeof(char) * f->reclen) + 5);
  /* File IO key buffers */ 
  f->kbuf1 = (char *)malloc((sizeof(char) * f->reclen) + 5);
  f->kbuf2 = (char *)malloc((sizeof(char) * f->reclen) + 5);

#ifdef DEBUG_FILEDBIO_RTS0
  fprintf(stderr, "debug : tcob_open_idx : 0080.140 : r=%d\n", r); 
#endif                                       

  return r;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                           tcob_close_idx                                |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_close_idx(struct file_desc *f, char *record) 
{
  int r=0, fdbp;
  DB *fdbpx;

  /* Close BDB index file */
  fdbpx = (DB *)f->dbpx;
  fdbpx->close(fdbpx);
  f->dbpx = NULL;

  /* Close sequential data file */
  fdbp = (int)f->dbp;
  close(fdbp);
  f->dbp = NULL;

  if (f->sbuf1 != NULL) 
     free(f->sbuf1);  
  if (f->kbuf1 != NULL) 
     free(f->kbuf1);  
  if (f->kbuf2 != NULL) 
     free(f->kbuf2);
  
#ifdef DEBUG_FILEDBIO_RTS0
  fprintf(stderr, "debug : tcob_close_idx : 0180.140 : r=%d\n", r); 
#endif                                       
  return r;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                       tcob_read_idx                                  |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_read_idx(struct file_desc *f, char *record, struct fld_desc *reclen_desc, char *reclen_buf, 
                  struct fld_desc *fkey, char *keybuf) 
{
  int result, keyidx, fdbp, r=0;
  unsigned int flags=0;
  DBT key, data, sdata;
  DB *fdbpx;
  struct altkey_desc *akd;        
  struct _clusterRecno rkey;

  f->flags.read_done = 0;

#ifdef DEBUG_FILEDBIO_RTS1
  fprintf(stderr, "debug : tcob_read_idx 0200.020 : f->rec_index=%d\n", f->rec_index);
#endif                 

  /* 
   Check to see if file is open. If not return File Status 47
   In accordance with the COBOL 85 Standard. 
  */
  if (f->dbp == NULL) {
     if (f->flags.optional && f->flags.file_missing) 
         RETURN_STATUS(10);        
     RETURN_STATUS(47);
  }

  /* Check to see that the record length is valid */
  if (f->reclen == -1)
      RETURN_STATUS(99);
  
  /* Check the mode the file is opened in to make sure that read
     is Allowed */
  if (((f->open_mode != FMOD_INPUT) && (f->open_mode != FMOD_IO)))
     RETURN_STATUS(47);
  
  /* Check that we didn't already hit eof */
  if ((f->open_mode == FMOD_INPUT || f->open_mode == FMOD_IO) && f->flags.eof_hit)
     RETURN_STATUS(46);
  
#ifdef DEBUG_FILEDBIO_RTS1
  fprintf(stderr, "debug : tcob_read_idx 0200.040 : f->reclen=%d;\n", f->reclen);
#endif                                       

  /* If there is a start record outstanding use it to fulfill the read */
  if (f->flags.start_record == 1) {

     r = tcob_read_rec(f, record, f->recloc, reclen_desc, reclen_buf); 

     f->flags.start_record = 0; 
     f->recloc = -1;
     f->flags.read_done = 1;

#ifdef DEBUG_FILEDBIO_RTS1
     fprintf(stderr, "debug : tcob_read_idx 0200.050 : r=%d\n", r);
#endif                 

     RETURN_STATUS(r);        
  }          

  keyidx = 0;
  if (fkey != NULL) {

#ifdef DEBUG_FILEDBIO_RTS1
//fkey, char *keybuf
  fprintf(stderr, "debug : tcob_read_idx 0200.060 : fkey->len=%d;\n", fkey->len);
  {
//     int i, ilen=f->reclen;
     int i, ilen=fkey->len;
//     fprintf(stderr, "debug : tcob_read_idx 0200.065 : record(%d)=", ilen);
     fprintf(stderr, "debug : tcob_read_idx 0200.065 : keybuf(%d)=", ilen);
     for (i=0; i<ilen; i++) {      
//        fprintf(stderr, "%c", *(record + i));
        fprintf(stderr, "%c", *(keybuf + i));
     }
     fprintf(stderr, ";\n");
  }
#endif

     if (f->ixd_desc->pic == fkey->pic) {
//        f->key_in_use = f->ixd_desc; 
        f->key_in_use = NULL; 
     }
     else {
        keyidx = 1;
        for (akd = (struct altkey_desc *)(f+1); akd->offset != -1; akd++) {
           if (akd->descriptor->pic == fkey->pic) {           
              f->key_in_use = akd; 
#ifdef DEBUG_FILEDBIO_RTS1
              fprintf(stderr, "debug : tcob_read_idx 0200.080 : keyidx=%d, akd->offset=%d, akd->descriptor->len=%d;\n", keyidx, akd->offset, akd->descriptor->len);
#endif                                       
              break;
           }
           keyidx++;
        }
        /* If the KEY option is used but key not found, then it is an error */
        if (akd->offset == -1) {
#ifdef DEBUG_FILEDBIO_RTS1
           fprintf(stderr, "debug : tcob_read_idx 0200.085 : akd->offset=%d, r=%d;\n", akd->offset, 99);
#endif                                       
           RETURN_STATUS(99);        
        }
     }
  }
  
  fdbpx = (DB *)f->dbpx;
  /* Use primary key to retrieve location and length of record */
  if (keyidx == 0) {
     sdata.data = record+f->rec_index;
     sdata.size = f->ixd_desc->len;        
     strncpy(f->kbuf1, "00", 2); 
     memcpy(f->kbuf1 + 2, sdata.data, sdata.size);
     key.data = f->kbuf1;
     key.size = sdata.size + 2;
//     result = fdbpx->seq(fdbpx, &key, &data, flags);
     result = fdbpx->get(fdbpx, &key, &data, flags);
  }
  else {
     /* Use secondary key to retrieve location and length of record  */
     sdata.data = keybuf;
//     sdata.size = f->ixd_desc->len;        
     sdata.size = akd->descriptor->len;        
     sprintf(f->kbuf1, "%02d", keyidx); 
//     memcpy(f->kbuf1 + 2, record+f->rec_index, f->ixd_desc->len);
     memcpy(f->kbuf1 + 2, sdata.data, sdata.size);
     key.data = f->kbuf1;
//     key.size = f->ixd_desc->len + 2;
     key.size = sdata.size + 2;
     if (akd->duplicates == 1) 
        result = fdbpx->seq(fdbpx, &key, &data, R_CURSOR);
     else
        result = fdbpx->get(fdbpx, &key, &data, flags);

  }
#ifdef DEBUG_FILEDBIO_RTS1
  {
   int i, ilen=key.size;
   fprintf(stderr, "debug : tcob_read_idx 0200.120 : keyidx=%d, f->kbuf1=", keyidx);
   for (i=0; i<ilen; i++) {      
      fprintf(stderr, "%c", f->kbuf1[i]);
   }
   fprintf(stderr, ", result=%d;\n", result);
  }
#endif                                       
  if (result != 0)
     RETURN_STATUS(23);

  /* Use location and length of record to retrieve data */
  fdbp = (int)f->dbp;
  rkey = *(struct _clusterRecno *)(data.data);
#ifdef DEBUG_FILEDBIO_RTS1
  fprintf(stderr, "debug : tcob_read_idx 0200.140 : rkey.start=%d, rkey.len=%d;\n", rkey.start , rkey.len);
#endif                                       
  if (rkey.len != f->reclen)
     RETURN_STATUS(99);

  r = tcob_read_rec(f, record, rkey.start, reclen_desc, reclen_buf); 
  
#ifdef DEBUG_FILEDBIO_RTS1
  fprintf(stderr, "debug : tcob_read_idx 0200.300 : r=%d;\n", r);
#endif                                       

  if (r == 0) {
     f->flags.read_done = 1;
#ifdef DEBUG_FILEDBIO_RTS1
    {
     int i, ilen=f->reclen;
     fprintf(stderr, "debug : tcob_read_idx 0200.340 : record(%d)=", ilen);
     for (i=0; i<ilen; i++) {      
        fprintf(stderr, "%c", *(record + i));
     }
     fprintf(stderr, ";\n");
    }
#endif
  }                                       

  RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                       tcob_write_idx                                   |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_write_idx(struct file_desc *f, char *record, struct fld_desc *reclen_desc, char *reclen_buf) 
{
  int result, fdbp, keyidx, keylen, flags=0, r=0;
  DBT key, data, skey, sdata;
  struct altkey_desc *akd;        
  struct _clusterRecno rkey;
  DB *fdbpx;
  
  f->flags.read_done = 0;
  f->flags.start_record = 0; 
  
  /* 
   Check to see if file is open. If not return File Status 48
   In accordance with the COBOL 85 Standard. 
  */
  if (f->dbp == NULL)
     RETURN_STATUS(48);
             
#ifdef DEBUG_FILEDBIO_RTS2
     fprintf(stderr, "debug : tcob_write_idx 0320.040 : f->reclen=%d\n", f->reclen);
#endif

  /* Check to see that the record length is valid */
  if (f->reclen == -1) 
     RETURN_STATUS(99);
  
  /* Check the file open mode to ensure a write is permitted */
  if (((f->open_mode != FMOD_OUTPUT) && 
       (f->open_mode != FMOD_IO) && 
       (f->open_mode != FMOD_EXTEND)))
     RETURN_STATUS(48);

  fdbpx = (DB *)f->dbpx;
  fdbp  = (int)f->dbp;

  /* Check if the primary key exists */
  strncpy(f->kbuf1, "00", 2); 
  memcpy(f->kbuf1 + 2, record+f->rec_index, f->ixd_desc->len);
  key.data = f->kbuf1;
  key.size = f->ixd_desc->len + 2;
  result = fdbpx->get(fdbpx, &key, &data, flags);
#ifdef DEBUG_FILEDBIO_RTS2
  fprintf(stderr, "debug : tcob_write_idx 0320.120 : result=%d, f->ixd_desc->len=%d\n", result, f->ixd_desc->len);
#endif                                       
  if (result == 0)
     RETURN_STATUS(22);
  if (result == -1) 
     RETURN_STATUS(99);

  /* Check if any secondary key(s) exist */
  keyidx = 1;
  for (akd = (struct altkey_desc *)(f+1); akd->offset != -1; akd++) {
      sprintf(f->kbuf1, "%02d", keyidx);
      keylen =  akd->descriptor->len + 2;
      memcpy(f->kbuf1 + 2, record+akd->offset, akd->descriptor->len);
      /* Add primary key suffix to secondary keys with duplicates */
      if (akd->duplicates == 1) {
         memcpy(f->kbuf1 + keylen, record+f->rec_index, f->ixd_desc->len);
         keylen = keylen + f->ixd_desc->len;
      }
      skey.data  = f->kbuf1;
      skey.size  = keylen;
      result = fdbpx->get(fdbpx, &skey, &sdata, flags);
#ifdef DEBUG_FILEDBIO_RTS2
      fprintf(stderr, "debug : tcob_write_idx 0320.130 : result=%d\n", result);
#endif                                       
      if (result == 0) { 
         r = 22;
         break;
      }
#ifdef DEBUG_FILEDBIO_RTS2
      {
       int i, ilen=keylen;
       fprintf(stderr, "debug : tcob_write_idx 0320.140 : keyidx=%d, f->kbuf1=", keyidx);
       for (i=0; i<ilen; i++) {      
          fprintf(stderr, "%c", f->kbuf1[i]);
       }
       fprintf(stderr, ";\n");
      }
#endif                                       
      keyidx++;
  }

  if (r != 0)
     RETURN_STATUS(r);

  result = lseek(fdbp, 0, SEEK_END);     
#ifdef DEBUG_FILEDBIO_RTS2
  fprintf(stderr, "debug : tcob_write_idx 0320.150 : result=%d\n", result);
#endif                                       
  if (result == -1) 
     RETURN_STATUS(99);

  flags = R_NOOVERWRITE;          

  /* Write primary key */
  strncpy(f->kbuf1, "00", 2); 
  memcpy(f->kbuf1 + 2, record+f->rec_index, f->ixd_desc->len);
  key.data = f->kbuf1;
  key.size = f->ixd_desc->len + 2;
  rkey.start = result; 
  rkey.len   = f->reclen; 
  data.data  = &rkey;
  data.size  = sizeof(struct _clusterRecno);
  result = fdbpx->put(fdbpx, &key, &data, flags);
#ifdef DEBUG_FILEDBIO_RTS2
  fprintf(stderr, "debug : tcob_write_idx 0320.160 : result=%d\n", result);
#endif                                       
  if (result == 1)
     RETURN_STATUS(22);
  if (result == -1) 
     RETURN_STATUS(99);

#ifdef DEBUG_FILEDBIO_RTS2
  {
   fprintf(stderr, "debug : tcob_write_idx 0320.170 : f->ixd_desc->len=%d\n", f->ixd_desc->len);
   int i, ilen=f->ixd_desc->len + 2;
   fprintf(stderr, "debug : tcob_write_idx 0320.175 : f->kbuf1=");
   for (i=0; i<ilen; i++) {      
      fprintf(stderr, "%c", f->kbuf1[i]);
   }
   fprintf(stderr, ";\n");
   fprintf(stderr, "debug : tcob_write_idx 0320.180 : rkey.start=%d, rkey.len=%d;\n", rkey.start , rkey.len);
  }
#endif                                       

  /* Write secondary key(s) */
  keyidx = 1;
  for (akd = (struct altkey_desc *)(f+1); akd->offset != -1; akd++) {
      sprintf(f->kbuf1, "%02d", keyidx);
      keylen =  akd->descriptor->len + 2;
//      memcpy(f->kbuf1 + 2, record+akd->offset, akd->descriptor->len);
      memcpy(f->kbuf1 + 2, record+akd->offset, akd->descriptor->len);
      /* Add primary key suffix to secondary keys with duplicates */
      if (akd->duplicates == 1) {
         memcpy(f->kbuf1 + keylen, record+f->rec_index, f->ixd_desc->len);
         keylen = keylen + f->ixd_desc->len;
      }
      skey.data  = f->kbuf1;
//      skey.size  = akd->descriptor->len + 2;
      skey.size  = keylen;
      /* sdata.data = record+f->rec_index; */
      /* sdata.size = f->ixd_desc->len; */
      sdata.data  = &rkey;
      sdata.size  = sizeof(struct _clusterRecno);
      result = fdbpx->put(fdbpx, &skey, &sdata, flags);
#ifdef DEBUG_FILEDBIO_RTS2
      fprintf(stderr, "debug : tcob_write_idx 0320.200 : result=%d\n", result);
#endif                                       
      if (result != 0) { 
         r = 99;
         break;
      }
#ifdef DEBUG_FILEDBIO_RTS2
      {
//       int i, ilen=akd->descriptor->len + 2;
       int i, ilen=keylen;
       fprintf(stderr, "debug : tcob_write_idx 0320.240 : keyidx=%d, f->kbuf1=", keyidx);
       for (i=0; i<ilen; i++) {      
          fprintf(stderr, "%c", f->kbuf1[i]);
       }
       fprintf(stderr, ";\n");
      }
#endif                                       
      keyidx++;
  }

  if (r != 0)
     RETURN_STATUS(r);

#ifdef USE_SYNC
  result = fdbpx->sync(fdbpx, 0);
#endif

  r = tcob_write_rec(f, record, rkey.start, reclen_desc, reclen_buf); 

#ifdef DEBUG_FILEDBIO_RTS2
  fprintf(stderr, "debug : tcob_write_idx 0320.300 : r=%d;\n", r);
#endif                                       

  RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                       tcob_rewrite_idx                                 |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_rewrite_idx(struct file_desc *f, char *record, struct fld_desc *reclen_desc, char *reclen_buf) 
{
  int result, keyidx, flags=0, keylen, r=0;
  DBT key, data, sdata, skey1, skey2;
  struct altkey_desc *akd;
  struct _clusterRecno rkey, rkey1;
  DB *fdbpx;

  f->flags.read_done = 0;
  f->flags.start_record = 0; 

  /* 
   Check to see if file is open. If not return File Status 49
   In accordance with the COBOL 85 Standard. 
  */
  if (f->dbp == NULL)
     RETURN_STATUS(49);
  
  /* Check to see that the record length is valid */
  if (f->reclen == -1)
     RETURN_STATUS(99);
  
  /* Check the mode the file is opened in to make sure that rewrite  is Allowed */
  if (f->open_mode != FMOD_IO)
     RETURN_STATUS(49);
  
  /* Check that we didn't already hit eof */
  if ((f->open_mode == FMOD_IO) && f->flags.eof_hit)
     RETURN_STATUS(43);
  
  /* Check that we had a previous successful read */
//  if (!read_done)
//     RETURN_STATUS(43);
  
  fdbpx = (DB *)f->dbpx;
//  fdbp  = (int)f->dbp;
 
  /* Check if the primary key exists */
  keyidx = 0;
  keylen = f->ixd_desc->len + 2;
  strncpy(f->kbuf1, "00", 2); 
  memcpy(f->kbuf1 + 2, record+f->rec_index, f->ixd_desc->len);
  key.data = f->kbuf1;
  key.size = f->ixd_desc->len + 2;
  result = fdbpx->get(fdbpx, &key, &data, flags);
#ifdef DEBUG_FILEDBIO_RTS6
  fprintf(stderr, "debug : tcob_rewrite_idx 0600.080 : result=%d, keyidx=%d, keylen=%d;\n", result, keyidx, keylen);
  {
   int i, ilen=keylen;
   fprintf(stderr, "debug : tcob_rewrite_idx 0600.085 : f->kbuf1(%d)=", keylen);
   for (i=0; i<ilen; i++) {      
      fprintf(stderr, "%c", f->kbuf1[i]);
   }
   fprintf(stderr, ";\n");
  }
#endif                                       
  if (result == 1)
     RETURN_STATUS(22);
  if (result == -1) 
     RETURN_STATUS(99);

  /* Use location and length of record to retrieve data */
  rkey = *(struct _clusterRecno *)(data.data);
  if (rkey.len != f->reclen)
      RETURN_STATUS(99);

  /* Verify the secondary key(s) - If key exists and duplicates are permitted */
  keyidx = 1;
  r = 0;
  for (akd = (struct altkey_desc *)(f+1); akd->offset != -1; akd++) {
     skey1.data = record+akd->offset;
     skey1.size = akd->descriptor->len;
     keylen = akd->descriptor->len + 2;
     sprintf(f->kbuf2, "%02d", keyidx); 
     memcpy(f->kbuf2 + 2, record+akd->offset, akd->descriptor->len);
     if (akd->duplicates == 1) {
        memcpy(f->kbuf2 + keylen, record+f->rec_index, f->ixd_desc->len);
        keylen = keylen + f->ixd_desc->len;
     }
     skey1.data = f->kbuf2; 
     skey1.size = keylen;
     result = fdbpx->get(fdbpx, &skey1, &data, flags);
     /* If NO secondary key duplicates permitted reset CURSOR */
     if (result == 0) {
        /* Use location and length of record to retrieve data */
        rkey1 = *(struct _clusterRecno *)(data.data);
        if (rkey1.start != rkey.start)
           r++;
     }
     keyidx++;
  }
  /* Reset the primary key to CURSOR */
  keylen = f->ixd_desc->len + 2;
  strncpy(f->kbuf1, "00", 2); 
  memcpy(f->kbuf1 + 2, record+f->rec_index, f->ixd_desc->len);
  key.data = f->kbuf1;
  key.size = f->ixd_desc->len + 2;
  result = fdbpx->get(fdbpx, &key, &data, flags);

  /* If NO secondary key duplicates permitted reset CURSOR */
  if (r != 0) 
     RETURN_STATUS(22);

  r = tcob_read_rec(f, f->sbuf1, rkey.start, reclen_desc, reclen_buf); 
#ifdef DEBUG_FILEDBIO_RTS6
  fprintf(stderr, "debug : tcob_rewrite_idx 0600.120 : r=%d\n", r);
#endif                                       
  if (r != 0) 
     RETURN_STATUS(r);

#ifdef DEBUG_FILEDBIO_RTS6
  {
   int i, ilen=f->reclen;
   fprintf(stderr, "debug : tcob_rewrite_idx 0600.140 : f->sbuf1(%d)=", ilen);
   for (i=0; i<ilen; i++) {      
      fprintf(stderr, "%c", f->sbuf1[i]);
   }
   fprintf(stderr, ";\n");
  }
#endif                                       

  // Compare secondary keys to determine if changes are required 
  keyidx = 1;
  for (akd = (struct altkey_desc *)(f+1); akd->offset != -1; akd++) {
     skey1.data = record+akd->offset;
     skey1.size = akd->descriptor->len;
     skey2.data = f->sbuf1+akd->offset;
     skey2.size = akd->descriptor->len;
     // Change secondary keys as required 
     if (memcmp(skey1.data, skey2.data, skey1.size) != 0) {
        // Delete old secondary key 
        sprintf(f->kbuf2, "%02d", keyidx); 
//        memcpy(f->kbuf2 + 2, f->sbuf1+akd->offset, akd->descriptor->len);
        memcpy(f->kbuf2 + 2, f->sbuf1+akd->offset, akd->descriptor->len);
        keylen = akd->descriptor->len + 2;
        if (akd->duplicates == 1) {
           memcpy(f->kbuf2 + keylen, record+f->rec_index, f->ixd_desc->len);
           keylen = keylen + f->ixd_desc->len;
        }
        skey1.data = f->kbuf2; 
//        skey1.size = akd->descriptor->len + 2;
        skey1.size = keylen;
        result = fdbpx->del(fdbpx, &skey1, flags);
#ifdef DEBUG_FILEDBIO_RTS6
        fprintf(stderr, "debug : tcob_rewrite_idx 0600.240 : result=%d, keyidx=%d;\n", result, keyidx);
        {
         int i, ilen=keylen;
         fprintf(stderr, "debug : tcob_rewrite_idx 0600.245 : f->kbuf2(%d)=", ilen);
         for (i=0; i<ilen; i++) {      
           fprintf(stderr, "%c", f->kbuf2[i]);
         }
         fprintf(stderr, ";\n");
        }
#endif                                       
        if (result != 0) { 
           r = 99;
           break;
        }
        // Write new secondary key 
        sprintf(f->kbuf2, "%02d", keyidx); 
        memcpy(f->kbuf2 + 2, record+akd->offset, akd->descriptor->len);
        keylen = akd->descriptor->len + 2;
        if (akd->duplicates == 1) {
           memcpy(f->kbuf2 + keylen, record+f->rec_index, f->ixd_desc->len);
           keylen = keylen + f->ixd_desc->len;
        }
        skey2.data = f->kbuf2; 
//        skey2.size = akd->descriptor->len + 2;
        skey2.size = keylen;
        sdata.data = &rkey;
        sdata.size = sizeof(struct _clusterRecno);
        result = fdbpx->put(fdbpx, &skey2, &sdata, flags);
#ifdef DEBUG_FILEDBIO_RTS6
        fprintf(stderr, "debug : tcob_rewrite_idx 0600.280 : result=%d\n", result);
        {
         int i, ilen=keylen;
         fprintf(stderr, "debug : tcob_rewrite_idx 0600.285 : f->kbuf2(%d)=", ilen);
         for (i=0; i<ilen; i++) {      
           fprintf(stderr, "%c", f->kbuf2[i]);
         }
         fprintf(stderr, ";\n");
        }
#endif                                       
        if (result != 0) { 
           r = 99;
           break;
        }
     }
     keyidx++;
  }

#ifdef DEBUG_FILEDBIO_RTS6
  fprintf(stderr, "debug : tcob_rewrite_idx 0600.320 : r=%d\n", r);
#endif                                       
  if (r != 0) 
     RETURN_STATUS(r);

  r = tcob_write_rec(f, record, rkey.start, reclen_desc, reclen_buf); 
#ifdef DEBUG_FILEDBIO_RTS6
  fprintf(stderr, "debug : tcob_rewrite_idx 0600.360 : r=%d\n", r);
#endif                                       

  RETURN_STATUS(r);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                       tcob_read_next_idx                               |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_read_next_idx(struct file_desc *f, char *record,  struct fld_desc *reclen_desc, char *reclen_buf) 
{
  int result, fdbp, keyidx, r=0;
  int flags=0;
//  DBT key, data, sdata, skey;
  DBT key, data, sdata;
  struct _clusterRecno rkey;
  struct altkey_desc *akd;
  DB *fdbpx;
  

  f->flags.read_done = 0;

#ifdef DEBUG_FILEDBIO_RTS1
  fprintf(stderr, "debug : tcob_read_next_idx 0883.020 : f->reclen=%d;\n", f->reclen);
#endif                                       
  /* 
    Check to see if file is open. If not return File Status 47
    In accordance with the COBOL-85 standard. 
  */
  if (f->dbp == NULL) {
     if (f->flags.optional && f->flags.file_missing) 
        RETURN_STATUS(10);        
     RETURN_STATUS(47);
  }
  
  /* Check the mode the file is opened in to make sure that read is Allowed */
  if (((f->open_mode != FMOD_INPUT) && (f->open_mode != FMOD_IO)))
     RETURN_STATUS(47);
  
  /* Check that we didn't already hit eof */
  if ((f->open_mode == FMOD_INPUT || f->open_mode == FMOD_IO) && f->flags.eof_hit)
     RETURN_STATUS(46);
  
#ifdef DEBUG_FILEDBIO_RTS1
  fprintf(stderr, "debug : tcob_read_next_idx 0883.040 : f->recloc=%d;\n", f->recloc);
#endif                                       

  /* If there is a start record outstanding use it to fulfill the read */
  if (f->flags.start_record == 1) {

     r = tcob_read_rec(f, record, f->recloc, reclen_desc, reclen_buf); 

#ifdef DEBUG_FILEDBIO_RTS1
     fprintf(stderr, "debug : tcob_read_next_idx 0883.050 : r=%d\n", r);
#endif                 
     if (r == 0) {
#ifdef DEBUG_FILEDBIO_RTS1
        {
         int i, ilen=f->reclen;
         fprintf(stderr, "debug : tcob_read_next_idx 0883.055 : record(%d)=", ilen);
         for (i=0; i<ilen; i++) {      
            fprintf(stderr, "%c", *(record + i));
         }
         fprintf(stderr, ";\n");
        }
#endif
        f->flags.start_record = 0; 
        f->recloc = -1;
        f->flags.read_done = 1;
    }                                       

     RETURN_STATUS(r);        
  }        

#ifdef DEBUG_FILEDBIO_RTS1
  fprintf(stderr, "debug : tcob_read_next_idx 0883.060 : f->reclen=%d;\n", f->reclen);
#endif                                       

  fdbpx = (DB *)f->dbpx;
  fdbp  = (int)f->dbp;
 
  /* Default selected key to primary key */
  keyidx = 0;

  /* Secondary key previously selected */
  if (f->key_in_use != NULL) {
     akd = f->key_in_use;
     keyidx = (int)akd->keyindex;
#ifdef DEBUG_FILEDBIO_RTS1
     fprintf(stderr, "debug : tcob_read_next_idx 0883.120 : keyidx=%d, akd->offset=%d, akd->descriptor->len=%d;\n", keyidx, akd->offset, akd->descriptor->len);
#endif                                       
  }

  if (f->flags.init_done == 1) {
    flags=R_CURSOR;
    f->flags.init_done = 0; 
  }
  else { 
    flags=R_NEXT;     
  }

  /* Use primary key to retrieve location and length of record */
  if (keyidx == 0) {
     sdata.data = record+f->rec_index;
//     sdata.data = f->start_record+f->rec_index;
     sdata.size = f->ixd_desc->len;        
     strncpy(f->kbuf1, "00", 2); 
     memcpy(f->kbuf1 + 2, sdata.data, sdata.size);
     key.data = f->kbuf1;
     key.size = sdata.size + 2;
#ifdef DEBUG_FILEDBIO_RTS1
  {
   int i, ilen=key.size;
   fprintf(stderr, "debug : tcob_read_next_idx 0883.130 : f->kbuf1(%d)=", ilen);
   for (i=0; i<ilen; i++) {      
      fprintf(stderr, "%c", *((char *)key.data + i));
   }
   fprintf(stderr, ";\n");
  }
#endif                                       
     result = fdbpx->seq(fdbpx, &key, &data, flags);
  }
  else {
     sprintf(f->kbuf1, "%02d", keyidx); 
//     sdata.data = keybuf;
     sdata.data = record+akd->offset;
//     sdata.data = f->start_record+akd->offset;
     sdata.size = akd->descriptor->len;        
//     skey.size = f->ixd_desc->len + 2;
     memcpy(f->kbuf1 + 2, sdata.data, sdata.size);
//     memcpy(f->kbuf1 + 2, record+f->rec_index, f->ixd_desc->len);
     /* result = fdbpx->seq(fdbpx, &skey, &sdata, flags); */
     key.data = f->kbuf1;
     key.size = sdata.size + 2;
#ifdef DEBUG_FILEDBIO_RTS1
  {
   int i, ilen=key.size;
   fprintf(stderr, "debug : tcob_read_next_idx 0883.140 : f->kbuf1(%d)=", ilen);
   for (i=0; i<ilen; i++) {      
      fprintf(stderr, "%c", *((char *)key.data + i));
   }
   fprintf(stderr, ";\n");
  }
#endif                                       
     result = fdbpx->seq(fdbpx, &key, &data, flags);
//     if (result != 0)
//        RETURN_STATUS(23);

  }
#ifdef DEBUG_FILEDBIO_RTS1
  {
   int i, ilen=key.size;
   fprintf(stderr, "debug : tcob_read_next_idx 0883.220 : result=%d, keyidx=%d, ilen=%d, key.data=", result, keyidx, ilen);
   for (i=0; i<ilen; i++) {      
      fprintf(stderr, "%c", *((char *)key.data + i));
   }
   fprintf(stderr, ";\n");
  }
#endif                                       

  /* Check for End-Of-Key within key-index */
  if (result != 0) {
     r = 1;
  }
  else {
     if ( (*(char *)key.data != f->kbuf1[0]) || ( (*(char *)(key.data + 1) != f->kbuf1[1]) ) ) 
        r = 1;
  }

  if (r != 0) {
     if (f->access_mode == 1) {
        RETURN_STATUS(10);
     }
     else {
        RETURN_STATUS(23);
     }
  }

  /* Use location and length of record to retrieve data */
  fdbp = (int)f->dbp;
  rkey = *(struct _clusterRecno *)(data.data);
#ifdef DEBUG_FILEDBIO_RTS1
  fprintf(stderr, "debug : tcob_read_next_idx 0883.240 : rkey.start=%d, rkey.len=%d, f->reclen=%d;\n", rkey.start , rkey.len, f->reclen);
#endif                                       
  if (rkey.len != f->reclen)
     RETURN_STATUS(99);

  r = tcob_read_rec(f, record, rkey.start, reclen_desc, reclen_buf); 

#ifdef DEBUG_FILEDBIO_RTS1
  fprintf(stderr, "debug : tcob_read_next_idx 0883.320 : r=%d, errno=%d;\n", r, errno);
#endif                                       

  if (r == 0) {
    f->flags.read_done = 1;
#ifdef DEBUG_FILEDBIO_RTS1
    {
     int i, ilen=f->reclen;
     fprintf(stderr, "debug : tcob_read_next_idx 0883.340 : record(%d)=", ilen);
     for (i=0; i<ilen; i++) {      
        fprintf(stderr, "%c", *(record + i));
     }
     fprintf(stderr, ";\n");
    }
#endif
  }                                       

  RETURN_STATUS(r);

}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                       tcob_start_idx                                   |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_start_idx(struct file_desc *f, char *record, int cond, struct fld_desc *recfdesc, char *key_ptr) 
{
  int result, rc, rcond, keyidx, keylen, r=0;
  int flags=R_CURSOR;
  DBT key, data, sdata;
  int alternate=0;
  DB *fdbpx;
  struct altkey_desc *akd;
  struct _clusterRecno rkey;
        
  f->flags.eof_hit = 0;  /* additon AL : reset eof if already positioned */
  f->flags.read_done = 0;

#ifdef DEBUG_FILEDBIO_RTS4
     fprintf(stderr, "debug : tcob_start_idx 0650.040 : \n");
#endif                                       

  /* 
   Check to see if file is open. 
   If not return File Status 47 in accordance with the COBOL 85 Standard. 
  */
  if (f->dbp == NULL)
     RETURN_STATUS(47);
  
#ifdef DEBUG_FILEDBIO_RTS4
     fprintf(stderr, "debug : tcob_start_idx 0650.050 : \n");
#endif                                       

  /* Check to see that the record length is valid */
  if (f->reclen == -1)
     RETURN_STATUS(99);
  
  /* 
   Check the mode the file is opened in to make sure that 
   start is Allowed 
  */
  if (((f->open_mode != FMOD_INPUT) && (f->open_mode != FMOD_IO)))
     RETURN_STATUS(47);
  
  rcond = cond;
#ifdef DEBUG_FILEDBIO_RTS4
     fprintf(stderr, "debug : tcob_start_idx 0650.060 : cond=%d;\n", cond);
#endif                                       

#ifdef DEBUG_FILEDBIO_RTS4
  if (recfdesc != NULL)     
     fprintf(stderr, "debug : tcob_start_idx 0650.070 : recfdesc=%d, key_ptr=%d;\n", recfdesc, key_ptr);
  else
     fprintf(stderr, "debug : tcob_start_idx 0650.070 : recfdesc=NULL, key_ptr=%d;\n", key_ptr);
#endif                                       

  if (recfdesc != NULL) {    
#ifdef DEBUG_FILEDBIO_RTS4
     fprintf(stderr, "debug : tcob_start_idx 0650.075 : recfdesc=%d;\n", recfdesc);
#endif                                       
    f->key_in_use = NULL;
    keyidx = 1;
    for (akd = (struct altkey_desc *)(f+1); akd->offset != -1; akd++) {
#ifdef DEBUG_FILEDBIO_RTS4
       fprintf(stderr, "debug : tcob_start_idx 0650.077 : keyidx=%d;\n", keyidx);
#endif                                       
       if (akd->descriptor == recfdesc) {
          f->key_in_use = akd;
          alternate = 1;
//          keyidx = 0;  
          break;
       }
       keyidx++;
    }
#ifdef DEBUG_FILEDBIO_RTS4
    fprintf(stderr, "debug : tcob_start_idx 0650.080 : key=%d, offset=%d, len=%d;\n", keyidx, akd->offset, akd->descriptor->len);
#endif                                       
  }
  else {
    /* When no key is given default to first record of primary key */
    rcond = -1;
  }  

  if (alternate == 0)
     keyidx = 0;

  fdbpx = (DB *)f->dbpx;

  if (keyidx == 0) {
     sdata.data = record+f->rec_index;
     sdata.size = f->ixd_desc->len;
  }
  else {
//     sdata.data = key_ptr;
     sdata.data = record+akd->offset;
     sdata.size = akd->descriptor->len;
  }
  sprintf(f->kbuf1, "%02d", keyidx); 
  memcpy(f->kbuf1 + 2, sdata.data, sdata.size);
  key.data = f->kbuf1;
  key.size = sdata.size + 2;
  keylen = sdata.size + 2;
  memset(f->sbuf1, 0, f->reclen);
  memcpy(f->sbuf1, f->kbuf1, key.size);

#ifdef DEBUG_FILEDBIO_RTS4
  {
   if (recfdesc != NULL) {    
    fprintf(stderr, "debug : tcob_start_idx 0650.100 : key=%d, offset=%d, len=%d;\n", keyidx, akd->offset, akd->descriptor->len);
    fprintf(stderr, "debug : tcob_start_idx 0650.120 : key.size=%d, f->reclen=%d\n", key.size, f->reclen);
//    int i, ilen=akd->descriptor->len + 2;
    int i, ilen=key.size;
    fprintf(stderr, "debug : tcob_start_idx 0650.125 : f->kbuf1(%d)=", ilen);
    for (i=0; i<ilen; i++) {      
        fprintf(stderr, "%c", f->kbuf1[i]);
    }
    fprintf(stderr, ";\n");
   }
  }
#endif                                       
   
  switch(rcond) 
  {
    case TCB_CONDITIONAL_EQUAL:  
       flags = R_CURSOR;
//       result = fdbpx->get(fdbpx, &key, &data, flags);
       result = fdbpx->seq(fdbpx, &key, &data, flags);
       if (result != 0)
          r = 23;
#ifdef DEBUG_FILEDBIO_RTS4
  {
    int i, ilen=key.size;
    fprintf(stderr, "debug : tcob_start_idx 0650.220 : f->sbuf1(%d)=", ilen);
    for (i=0; i<ilen; i++) {      
//        fprintf(stderr, "%c", *(f->sbuf1 + i));
        fprintf(stderr, "%c", *((char *)key.data + i));
    }
    fprintf(stderr, ";\n");
  }
#endif                                       
       rc = memcmp(key.data, f->sbuf1, keylen);
       if (rc != 0)
          r = 23;
       break;

    case TCB_CONDITIONAL_LT: 
       flags=R_CURSOR;
       result = fdbpx->seq(fdbpx, &key, &data, flags);
#ifdef DEBUG_FILEDBIO_RTS4
       fprintf(stderr, "debug : tcob_start_idx 0650.320 : result=%d\n", result);
#endif                                       
       if (result == 1) {
          flags=R_LAST;
          result = fdbpx->seq(fdbpx, &key, &data, flags);
#ifdef DEBUG_FILEDBIO_RTS4
       fprintf(stderr, "debug : tcob_start_idx 0650.325 : result=%d\n", result);
#endif                                       
          if (result != 0) 
             r = 23;
          else {
#ifdef DEBUG_FILEDBIO_RTS4
       {
       int i, ilen=key.size;
       fprintf(stderr, "debug : tcob_start_idx 0650.330 : key.data(%d)=", ilen);
       for (i=0; i<ilen; i++) {      
           fprintf(stderr, "%c", *((char *)key.data + i));
       }
       fprintf(stderr, ";\n");
       }
#endif                                       
             rc = memcmp(key.data, f->sbuf1, key.size);
#ifdef DEBUG_FILEDBIO_RTS4
       fprintf(stderr, "debug : tcob_start_idx 0650.335 : rc=%d\n", rc);
#endif                                       
             if (rc == 1) 
               r = 23;
          }
       }
       else {
          flags=R_PREV;
          result = fdbpx->seq(fdbpx, &key, &data, flags);       
#ifdef DEBUG_FILEDBIO_RTS4
          fprintf(stderr, "debug : tcob_start_idx 0650.340 : result=%d\n", result);
#endif                                       
          rc = 0;
          while ((result == 0) && (rc == 0)) {
            rc = memcmp(key.data, f->sbuf1, key.size);
            if (rc == 1)  {
               flags=R_PREV;
               result = fdbpx->seq(fdbpx, &key, &data, flags);       
            }
          }
       }
       break;

    case TCB_CONDITIONAL_LTEQ: 
       flags = R_CURSOR;
       result = fdbpx->seq(fdbpx, &key, &data, flags);
#ifdef DEBUG_FILEDBIO_RTS4
       fprintf(stderr, "debug : tcob_start_idx 0650.420 : result=%d\n", result);
#endif                                       
       if (result != TCB_BDB_ERROR) {
#ifdef DEBUG_FILEDBIO_RTS4
       {
       int i, ilen=key.size;
       fprintf(stderr, "debug : tcob_start_idx 0650.425 : key.data(%d)=", ilen);
       for (i=0; i<ilen; i++) {      
           fprintf(stderr, "%c", *((char *)key.data + i));
       }
       fprintf(stderr, ";\n");
       }
#endif                                       
          if (result == 1)  {
             flags=R_LAST;
             result = fdbpx->seq(fdbpx, &key, &data, flags);       
#ifdef DEBUG_FILEDBIO_RTS4
       fprintf(stderr, "debug : tcob_start_idx 0650.460 : result=%d\n", result);
#endif                                       
             if (result != 0) {
                r = 23;
             }
             else {
                rc = memcmp(key.data, f->sbuf1, key.size);
                if (rc == 1) 
                   r = 23;
             }
          }
          else {
             flags=R_PREV;
             result = fdbpx->seq(fdbpx, &key, &data, flags);       
             rc = 0;
             while ((result == 0) && (rc == 0)) {
               rc = memcmp(key.data, f->sbuf1, key.size);
               if (rc == 1)  {
                  flags=R_PREV;
                  result = fdbpx->seq(fdbpx, &key, &data, flags);       
               }
             }
          }
       }
       break;

    case TCB_CONDITIONAL_GT: 
       flags = R_CURSOR;
       result = fdbpx->seq(fdbpx, &key, &data, flags);
#ifdef DEBUG_FILEDBIO_RTS4
       fprintf(stderr, "debug : tcob_start_idx 0650.520 : result=%d, keylen=%d;\n", result, keylen);
#endif                                       
       if (result != TCB_BDB_ERROR) {
         if (result == 0)  {
#ifdef DEBUG_FILEDBIO_RTS4
         {
          int i, ilen=key.size;
          fprintf(stderr, "debug : tcob_start_idx 0650.525 : key.data(%d)=", ilen);
          for (i=0; i<ilen; i++) {      
              fprintf(stderr, "%c", *((char *)key.data + i));
          }
          fprintf(stderr, ";\n");
          rkey = *(struct _clusterRecno *)(data.data);
          fprintf(stderr, "debug : tcob_start_idx 0650.530 : rkey.start=%d, rkey.len=%d, f->reclen=%d;\n", rkey.start , rkey.len, f->reclen);
         }
#endif                                       
             rc = 0;
             /* Read next key until GT condition */
             while ((result == 0) && (rc == 0)) {
               /* Compare keys */
               rc = memcmp(key.data, f->sbuf1, keylen);
#ifdef DEBUG_FILEDBIO_RTS4
         {
         fprintf(stderr, "debug : tcob_start_idx 0650.540 : rc=%d, result=%d\n", rc, result);
         int i, ilen=key.size;
         fprintf(stderr, "debug : tcob_start_idx 0650.545 : key.data(%d)=", ilen);
         for (i=0; i<ilen; i++) {      
             fprintf(stderr, "%c", *((char *)key.data + i));
         }
         fprintf(stderr, ";\n");
         rkey = *(struct _clusterRecno *)(data.data);
         fprintf(stderr, "debug : tcob_start_idx 0650.550 : rkey.start=%d, rkey.len=%d, f->reclen=%d;\n", rkey.start , rkey.len, f->reclen);
         }
#endif                                       
               if (rc == 0)  {
                  flags=R_NEXT;
                  result = fdbpx->seq(fdbpx, &key, &data, flags);       
               }
               else {
                  /* Check key index */
                  if (rc == 1)  {
                     int rc1;
//                     rc1 = memcmp(key.data, f->sbuf1, keylen - 2);               
                     rc1 = memcmp(key.data, f->sbuf1, 2);               
                      if (rc1 != 0) 
                        result = 1;
                  }
                  else {
                     result = 1;
                  }
               }
             }
         }

         if (result == 1) {
#ifdef DEBUG_FILEDBIO_RTS4
             fprintf(stderr, "debug : tcob_start_idx 0650.560 : result=%d\n", result);
#endif                                       
              r = 23;
         }
       }
       break;

    case TCB_CONDITIONAL_GTEQ:
#ifdef DEBUG_FILEDBIO_RTS4
     fprintf(stderr, "debug : tcob_start_idx 0650.620 : r=%d\n", r);
#endif                                       
//       flags = 0;
       flags = R_CURSOR;
       result = fdbpx->seq(fdbpx, &key, &data, flags);
#ifdef DEBUG_FILEDBIO_RTS4
     fprintf(stderr, "debug : tcob_start_idx 0650.540 : result=%d\n", result);
#endif                                       
       if (result != 0) 
          r = 23;
       break;

    /* When no key is given default to first record of primary key */
    case -1:
       flags = R_FIRST;
       result = fdbpx->seq(fdbpx, &key, &data, flags);
       break;

    case TCB_CONDITIONAL_NOTEQ:
    default:
       r = 99;
       break;
  }

  if (result == -1) 
     r = 99;
#ifdef DEBUG_FILEDBIO_RTS4
     fprintf(stderr, "debug : tcob_start_idx 0650.720 : r=%d\n", r);
#endif                                       

  if (r != 0)
     RETURN_STATUS(r);

  /* Use location and length of record to retrieve data */
  rkey = *(struct _clusterRecno *)(data.data);
#ifdef DEBUG_FILEDBIO_RTS4
  fprintf(stderr, "debug : tcob_start_idx 0650.740 : rkey.start=%d, rkey.len=%d, f->reclen=%d;\n", rkey.start , rkey.len, f->reclen);
#endif                                       
  if (rkey.len != f->reclen)
     RETURN_STATUS(99);

  f->recloc = rkey.start;
  f->flags.start_record = 1;

  RETURN_STATUS(r);                 
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                       tcob_delete_idx                                  |
 |                                                                        |
\*------------------------------------------------------------------------*/
int tcob_delete_idx(struct file_desc *f, char *record) 
{
  int result, ndel, nerr=0, keyidx, keylen, r=0;
  int flags=0;
  DB *fdbpx;
  DBT key, data;
  struct _clusterRecno prkey;
  struct altkey_desc *akd;
        
#ifdef DEBUG_FILEDBIO_RTS5
  fprintf(stderr, "debug : tcob_delete_idx 1590.020 : \n");
#endif                                       
  /* 
   Check to see if file is open. 
   If not return File Status 49 in accordance with the COBOL-85 standard. 
  */

  if (f->dbp == NULL)
     RETURN_STATUS(49);
  
  /* Check to see that the record length is valid */
  if (f->reclen == -1)
     RETURN_STATUS(99);
  
  
  /* Check the mode the file is opened in to make sure that delete
     is Allowed */
  if (f->open_mode != FMOD_IO)
     RETURN_STATUS(49);
  
  fdbpx = (DB *)f->dbpx;
  keyidx = 0;

#ifdef DEBUG_FILEDBIO_RTS5
  {
   int i, ilen=f->reclen;
   fprintf(stderr, "debug : tcob_delete_idx 1590.030 : record(%d)=", ilen);
   for (i=0; i<ilen; i++) {      
       fprintf(stderr, "%c", *(record + i));
   }
   fprintf(stderr, ";\n");
  }
#endif                                       

  /* Check if primary key exists */
  sprintf(f->kbuf1, "%02d", keyidx); 
  memcpy(f->kbuf1 + 2, record+f->rec_index, f->ixd_desc->len);
  key.data = f->kbuf1;
  key.size = f->ixd_desc->len + 2;

//  flags = R_CURSOR;
//  result = fdbpx->seq(fdbpx, &key, &data, flags);
  result = fdbpx->get(fdbpx, &key, &data, flags);
#ifdef DEBUG_FILEDBIO_RTS5
  fprintf(stderr, "debug : tcob_delete_idx 1590.040 : result=%d\n", result);
  {
   int i, ilen=key.size;
//   fprintf(stderr, "debug : tcob_delete_idx 1590.045 : key.data(%d)=", ilen);
   fprintf(stderr, "debug : tcob_delete_idx 1590.045 : f->kbuf1(%d)=", ilen);
   for (i=0; i<ilen; i++) {      
       fprintf(stderr, "%c", f->kbuf1[i]);
//       fprintf(stderr, "%c", *((char *)key.data + i));
   }
   fprintf(stderr, ";\n");
  }
#endif                                       
  if (result == 1)
     RETURN_STATUS(23);
  if (result != 0)
     RETURN_STATUS(99);
  prkey = *(struct _clusterRecno *)(data.data);

#ifdef DEBUG_FILEDBIO_RTS5
  fprintf(stderr, "debug : tcob_delete_idx 1590.060 : rkey.start=%d, rkey.len=%d, f->reclen=%d;\n", prkey.start , prkey.len, f->reclen);
#endif                                       
  if (prkey.len != f->reclen)
     RETURN_STATUS(99);

  /* Delete secondary keys */    
  keyidx = 1;
  for (akd = (struct altkey_desc *)(f+1); akd->offset != -1; akd++) {
     keylen =  akd->descriptor->len + 2;
     sprintf(f->kbuf1, "%02d", keyidx); 
     memcpy(f->kbuf1 + 2, record+akd->offset, akd->descriptor->len);
     /* Add primary key suffix to secondary key with duplicates to define unique key */
     if (akd->duplicates == 1) {
        memcpy(f->kbuf1 + keylen, record+f->rec_index, f->ixd_desc->len);
        keylen = keylen + f->ixd_desc->len;
     }
     key.data = f->kbuf1;
     key.size = keylen;
#ifdef DEBUG_FILEDBIO_RTS5
     {
      int i, ilen=keylen;
      fprintf(stderr, "debug : tcob_delete_idx 1590.075 : f->kbuf1(%d)=", ilen);
      for (i=0; i<ilen; i++) {      
         fprintf(stderr, "%c", *(f->kbuf1 + i));
      }
      fprintf(stderr, ";\n");
     }
#endif                                       
     result = fdbpx->del(fdbpx, &key, flags);       
#ifdef DEBUG_FILEDBIO_RTS5
     fprintf(stderr, "debug : tcob_delete_idx 1590.080 : result=%d\n", result);
#endif                                       
     /* If NOT ZERO, some sort of error */
     if (result == 0)
        ndel++;
     else
        nerr++;
     keyidx++;
  }

  /* If no errors occured the delete primary key */
  if (nerr == 0) {
     keyidx = 0;
     sprintf(f->kbuf1, "%02d", keyidx); 
     memcpy(f->kbuf1 + 2, record+f->rec_index, f->ixd_desc->len);
     key.data = f->kbuf1;
     key.size = f->ixd_desc->len + 2;
     result = fdbpx->del(fdbpx, &key, flags);       
     if (result == 0) {
        result = fdbpx->del(fdbpx, &key, flags);
#ifdef DEBUG_FILEDBIO_RTS5
        fprintf(stderr, "debug : tcob_delete_idx 1590.280 : result=%d\n", result);
        {
         int i, ilen=key.size;
         fprintf(stderr, "debug : tcob_delete_idx 1590.285 : f->kbuf1(%d)=", ilen);
         for (i=0; i<ilen; i++) {      
             fprintf(stderr, "%c", *(f->kbuf1 + i));
         }
         fprintf(stderr, ";\n");
        }
#endif                                       
        if (result != 0) 
           r = 99;
        /* else - should rewrite record data with blanks */
     }
  }
  else 
     r = 99;

  f->flags.init_done = 1; 

#ifdef DEBUG_FILEDBIO_RTS5
  fprintf(stderr, "debug : tcob_delete_idx 1590.300 : r=%d;\n", r);
#endif                                       

  RETURN_STATUS(r);
}

/* EOF fileiodb.c */
